# Copyright 2017, Rackspace US, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

import requests_mock
import webob.exc

from neutron.api import extensions
from neutron.common import config
from neutron.tests.unit.api.v2 import test_base
from neutron_lib import constants as n_constants
from neutron_lib import context
from neutron_lib import exceptions as lib_exc
from oslo_config import cfg
from oslo_utils import uuidutils

from neutron_lbaas.services.loadbalancer import constants as lb_const
from neutron_lbaas.services.loadbalancer import \
    proxy_plugin as loadbalancer_plugin
from neutron_lbaas.tests import base
from neutron_lbaas.tests.unit.db.loadbalancer import util

LB_PLUGIN_CLASS = (
    "neutron_lbaas.services.loadbalancer."
    "proxy_plugin.LoadBalancerProxyPluginv2"
)

_uuid = uuidutils.generate_uuid
_get_path = test_base._get_path


_subnet_id = "0c798ed8-33ba-11e2-8b28-000c291c4d14"
base_url = '{}/{}'.format('http://127.0.0.1:9876', 'v2.0/lbaas')


class TestLbaasProxyPluginDbTestCase(util.LbaasTestMixin,
                                     base.NeutronDbPluginV2TestCase):
    fmt = 'json'

    def setUp(self, core_plugin=None, lb_plugin=None, lbaas_provider=None,
              ext_mgr=None):
        service_plugins = {'lb_plugin_name': LB_PLUGIN_CLASS}

        # removing service-type because it resides in neutron and tests
        # dont care
        LBPlugin = loadbalancer_plugin.LoadBalancerProxyPluginv2
        sea_index = None
        for index, sea in enumerate(LBPlugin.supported_extension_aliases):
            if sea == 'service-type':
                sea_index = index
        if sea_index:
            del LBPlugin.supported_extension_aliases[sea_index]

        super(TestLbaasProxyPluginDbTestCase, self).setUp(
            ext_mgr=ext_mgr,
            service_plugins=service_plugins
        )

        if not ext_mgr:
            self.plugin = loadbalancer_plugin.LoadBalancerProxyPluginv2()
            # This is necessary because the automatic extension manager
            # finding algorithm below will find the loadbalancerv2
            # extension and fail to initizlize the main API router with
            # extensions' resources
            ext_mgr = util.ExtendedPluginAwareExtensionManager(
                LBPlugin.supported_extension_aliases)

            app = config.load_paste_app('extensions_test_app')
            self.ext_api = extensions.ExtensionMiddleware(app, ext_mgr=ext_mgr)

        self._subnet_id = _subnet_id
        # set quotas to -1 (unlimited)
        cfg.CONF.set_override('quota_loadbalancer', -1, group='QUOTAS')

    def _update_loadbalancer_api(self, lb_id, data):
        req = self.new_update_request('loadbalancers', data, lb_id)
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, req.get_response(self.ext_api))
        return resp, body

    def _delete_loadbalancer_api(self, lb_id):
        req = self.new_delete_request('loadbalancers', lb_id)
        resp = req.get_response(self.ext_api)
        return resp

    def _get_loadbalancer_api(self, lb_id):
        req = self.new_show_request('loadbalancers', lb_id)
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body

    def _list_loadbalancers_api(self):
        req = self.new_list_request('loadbalancers')
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body

    def _get_loadbalancer_stats_api(self, lb_id):
        req = self.new_show_request('loadbalancers', lb_id,
                                    subresource='stats')
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body

    def _get_loadbalancer_statuses_api(self, lb_id):
        req = self.new_show_request('loadbalancers', lb_id,
                                    subresource='statuses')
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body


class LbaasLoadBalancerTests(TestLbaasProxyPluginDbTestCase):
    url = '{}/{}'.format(base_url, 'loadbalancers')

    @requests_mock.mock()
    def test_create_loadbalancer(self, m, **extras):
        expected = {
            'name': 'vip1',
            'description': '',
            'admin_state_up': True,
            'provisioning_status': n_constants.ACTIVE,
            'operating_status': lb_const.ONLINE,
            'tenant_id': self._tenant_id,
            'listeners': [],
            'pools': [],
            'provider': 'lbaas'
        }

        res = expected.copy()
        res['id'] = 'fake_uuid'
        expected.update(extras)

        m.post(self.url, json={'loadbalancer': res})
        with self.subnet() as subnet:
            name = expected['name']
            with self.loadbalancer(name=name, subnet=subnet,
                                   no_delete=True, **extras) as lb:
                lb_id = lb['loadbalancer']['id']
                self.assertEqual('fake_uuid', lb_id)
                actual = dict((k, v)
                              for k, v in lb['loadbalancer'].items()
                              if k in expected)
                self.assertEqual(expected, actual)
            return lb

    @requests_mock.mock()
    def test_list_loadbalancers(self, m):
        name = 'lb_show'
        description = 'lb_show description'
        expected_values = {'name': name,
                           'description': description,
                           'vip_address': '10.0.0.10',
                           'admin_state_up': True,
                           'provisioning_status': n_constants.ACTIVE,
                           'operating_status': lb_const.ONLINE,
                           'listeners': [],
                           'provider': 'lbaas'}

        m.get(self.url, json={'loadbalancers': [expected_values]})
        resp, body = self._list_loadbalancers_api()
        self.assertEqual(1, len(body['loadbalancers']))
        for k in expected_values:
            self.assertEqual(expected_values[k],
                             body['loadbalancers'][0][k])

    def _build_lb_list(self):
        expected_values = []
        for name in ['lb1', 'lb2', 'lb3']:
            expected_values.append({'name': name,
                                    'description': 'lb_show description',
                                    'vip_address': '10.0.0.10',
                                    'admin_state_up': True,
                                    'provisioning_status': n_constants.ACTIVE,
                                    'operating_status': lb_const.ONLINE,
                                    'listeners': [],
                                    'provider': 'lbaas',
                                    'id': name})
        return expected_values

    @requests_mock.mock()
    def test_list_loadbalancers_with_sort_emulated(self, m):
        expected_values = self._build_lb_list()
        m.get(self.url, json={'loadbalancers': expected_values})

        self._test_list_with_sort(
            'loadbalancer',
            ({'loadbalancer': expected_values[0]},
             {'loadbalancer': expected_values[1]},
             {'loadbalancer': expected_values[2]}),
            [('name', 'asc')]
        )

    @requests_mock.mock()
    def test_list_loadbalancers_with_pagination_emulated(self, m):
        expected_values = self._build_lb_list()
        m.get(self.url, json={'loadbalancers': expected_values})
        self._test_list_with_pagination(
            'loadbalancer',
            ({'loadbalancer': expected_values[0]},
             {'loadbalancer': expected_values[1]},
             {'loadbalancer': expected_values[2]}),
            ('name', 'asc'), 2, 2
        )

    @requests_mock.mock()
    def test_list_loadbalancers_with_pagination_reverse_emulated(self, m):
        expected_values = self._build_lb_list()
        m.get(self.url, json={'loadbalancers': expected_values})
        self._test_list_with_pagination_reverse(
            'loadbalancer',
            ({'loadbalancer': expected_values[0]},
             {'loadbalancer': expected_values[1]},
             {'loadbalancer': expected_values[2]}),
            ('name', 'asc'), 2, 2
        )

    @requests_mock.mock()
    def test_show_loadbalancer(self, m):
        name = 'lb_show'
        description = 'lb_show description'
        lb_id = "testid"
        expected_values = {'name': name,
                           'description': description,
                           'vip_address': '10.0.0.10',
                           'admin_state_up': True,
                           'provisioning_status': n_constants.ACTIVE,
                           'operating_status': lb_const.ONLINE,
                           'listeners': [],
                           'provider': 'lbaas',
                           'id': lb_id}
        m.get("{}/{}".format(self.url, lb_id),
              json={'loadbalancer': expected_values})
        resp, body = self._get_loadbalancer_api(lb_id)
        for k in expected_values:
            self.assertEqual(expected_values[k],
                             body['loadbalancer'][k])

    @requests_mock.mock()
    def test_update_loadbalancer(self, m):
        loadbalancer_id = "test_uuid"
        name = 'new_loadbalancer'
        description = 'a crazy loadbalancer'
        expected_values = {'name': name,
                           'description': description,
                           'admin_state_up': False,
                           'provisioning_status': n_constants.ACTIVE,
                           'operating_status': lb_const.ONLINE,
                           'listeners': [],
                           'provider': 'lbaas',
                           'id': loadbalancer_id}

        m.put("{}/{}".format(self.url, loadbalancer_id),
              json={'loadbalancer': expected_values})
        # wonder why an update triggers a get...
        m.get("{}/{}".format(self.url, loadbalancer_id),
              json={'loadbalancer': expected_values})
        data = {'loadbalancer': {'name': name,
                                 'description': description,
                                 'admin_state_up': False}}
        resp, res = self._update_loadbalancer_api(loadbalancer_id,
                                                  data)
        for k in expected_values:
            self.assertEqual(expected_values[k],
                             res['loadbalancer'][k])

    @requests_mock.mock()
    def test_delete_loadbalancer(self, m):
        loadbalancer_id = "test_uuid"
        expected_values = {'name': "test",
                           'description': "test",
                           'vip_address': '10.0.0.10',
                           'admin_state_up': True,
                           'provisioning_status': n_constants.ACTIVE,
                           'operating_status': lb_const.ONLINE,
                           'listeners': [],
                           'provider': 'lbaas',
                           'id': loadbalancer_id}
        # wonder why an delete triggers a get...
        m.get("{}/{}".format(self.url, loadbalancer_id),
              json={'loadbalancer': expected_values})
        m.delete("{}/{}".format(self.url, loadbalancer_id))
        resp = self._delete_loadbalancer_api(loadbalancer_id)
        self.assertEqual(webob.exc.HTTPNoContent.code, resp.status_int)

    @requests_mock.mock()
    def test_delete_loadbalancer_when_loadbalancer_in_use(self, m):
        lb_id = "123"
        m.delete("{}/{}".format(self.url, lb_id), status_code=409)
        ctx = context.get_admin_context()
        # notice we raise a more generic exception then the stabdard plugin
        self.assertRaises(lib_exc.Conflict,
                          self.plugin.delete_loadbalancer,
                          ctx, lb_id)


class ListenerTestBase(TestLbaasProxyPluginDbTestCase):
    lb_id = uuid.uuid4().hex

    def _create_listener_api(self, data):
        req = self.new_create_request("listeners", data, self.fmt)
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body

    def _update_listener_api(self, listener_id, data):
        req = self.new_update_request('listeners', data, listener_id)
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, req.get_response(self.ext_api))
        return resp, body

    def _delete_listener_api(self, listener_id):
        req = self.new_delete_request('listeners', listener_id)
        resp = req.get_response(self.ext_api)
        return resp

    def _get_listener_api(self, listener_id):
        req = self.new_show_request('listeners', listener_id)
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body

    def _list_listeners_api(self):
        req = self.new_list_request('listeners')
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body


class LbaasListenerTests(ListenerTestBase):
    url = '{}/{}'.format(base_url, 'listeners')

    @requests_mock.mock()
    def test_create_listener(self, m):
        expected = {
            'protocol': 'HTTP',
            'protocol_port': 80,
            'admin_state_up': True,
            'tenant_id': self._tenant_id,
            'default_pool_id': None,
            'loadbalancers': [{'id': self.lb_id}],
            'id': '123'
        }

        m.post(self.url, json={'listener': expected})
        with self.listener(loadbalancer_id=self.lb_id,
                           no_delete=True) as listener:
            listener_id = listener['listener'].get('id')
            self.assertTrue(listener_id)
            actual = {}
            for k, v in listener['listener'].items():
                if k in expected:
                    actual[k] = v
            self.assertEqual(expected, actual)
        return listener

    @requests_mock.mock()
    def test_create_listener_same_port_same_load_balancer(self, m):
        m.post(self.url, status_code=409)
        self._create_listener(self.fmt, 'HTTP', 80,
                              loadbalancer_id=self.lb_id,
                              expected_res_status=409,
                              no_delete=True)

    @requests_mock.mock()
    def test_create_listener_with_tls_no_default_container(self, m, **extras):
        listener_data = {
            'protocol': lb_const.PROTOCOL_TERMINATED_HTTPS,
            'default_tls_container_ref': None,
            'protocol_port': 443,
            'admin_state_up': True,
            'tenant_id': self._tenant_id,
            'loadbalancer_id': self.lb_id,
        }
        m.post(self.url, status_code=400)
        listener_data.update(extras)
        self.assertRaises(
                        lib_exc.BadRequest,
                        self.plugin.create_listener,
                        context.get_admin_context(),
                        {'listener': listener_data})

    @requests_mock.mock()
    def test_create_listener_loadbalancer_id_does_not_exist(self, m):
        m.post(self.url, status_code=404)
        self._create_listener(self.fmt, 'HTTP', 80,
                              loadbalancer_id=uuidutils.generate_uuid(),
                              expected_res_status=404)

    @requests_mock.mock()
    def test_update_listener(self, m):
        name = 'new_listener'
        expected_values = {'name': name,
                           'protocol_port': 80,
                           'protocol': 'HTTP',
                           'connection_limit': 100,
                           'admin_state_up': False,
                           'tenant_id': self._tenant_id,
                           'loadbalancers': [{'id': self.lb_id}]}

        listener_id = uuidutils.generate_uuid()
        # needs a get before a put...
        m.get("{}/{}".format(self.url, listener_id),
              json={'listener': expected_values})
        m.put("{}/{}".format(self.url, listener_id),
              json={'listener': expected_values})
        data = {'listener': {'name': name,
                             'connection_limit': 100,
                             'admin_state_up': False}}
        resp, body = self._update_listener_api(listener_id, data)
        for k in expected_values:
            self.assertEqual(expected_values[k], body['listener'][k])

    @requests_mock.mock()
    def test_delete_listener(self, m):
        expected_values = {'name': 'test',
                           'protocol_port': 80,
                           'protocol': 'HTTP',
                           'connection_limit': 100,
                           'admin_state_up': False,
                           'tenant_id': self._tenant_id,
                           'loadbalancers': [{'id': self.lb_id}]}
        listener_id = uuidutils.generate_uuid()
        m.get("{}/{}".format(self.url, listener_id),
              json={'listener': expected_values})
        m.delete("{}/{}".format(self.url, listener_id))

        resp = self._delete_listener_api(listener_id)
        self.assertEqual(webob.exc.HTTPNoContent.code, resp.status_int)

    @requests_mock.mock()
    def test_show_listener(self, m):
        name = 'show_listener'
        expected_values = {'name': name,
                           'protocol_port': 80,
                           'protocol': 'HTTP',
                           'connection_limit': -1,
                           'admin_state_up': True,
                           'tenant_id': self._tenant_id,
                           'default_pool_id': None,
                           'loadbalancers': [{'id': self.lb_id}]}
        listener_id = uuidutils.generate_uuid()
        m.get("{}/{}".format(self.url, listener_id),
              json={'listener': expected_values})
        resp, body = self._get_listener_api(listener_id)
        for k in expected_values:
            self.assertEqual(expected_values[k], body['listener'][k])

    @requests_mock.mock()
    def test_list_listeners(self, m):
        name = 'list_listeners'
        expected_values = {'name': name,
                           'protocol_port': 80,
                           'protocol': 'HTTP',
                           'connection_limit': -1,
                           'admin_state_up': True,
                           'tenant_id': self._tenant_id,
                           'loadbalancers': [{'id': self.lb_id}]}

        listener_id = uuidutils.generate_uuid()
        m.get(self.url, json={'listeners': [expected_values]})
        expected_values['id'] = listener_id
        resp, body = self._list_listeners_api()
        listener_list = body['listeners']
        self.assertEqual(1, len(listener_list))
        for k in expected_values:
            self.assertEqual(expected_values[k], listener_list[0][k])


class LbaasL7Tests(ListenerTestBase):
    url = '{}/{}'.format(base_url, 'l7policies')

    def _rules(self, policy_id=uuidutils.generate_uuid()):
        return "{}/{}/rules".format(self.url, policy_id)

    @requests_mock.mock()
    def test_create_l7policy_invalid_listener_id(self, m, **extras):
        m.post(self.url, status_code=404)
        self._create_l7policy(self.fmt, uuidutils.generate_uuid(),
                              lb_const.L7_POLICY_ACTION_REJECT,
                              expected_res_status=webob.exc.HTTPNotFound.code)

    @requests_mock.mock()
    def test_create_l7policy(self, m, **extras):
        expected = {
            'action': lb_const.L7_POLICY_ACTION_REJECT,
            'redirect_pool_id': None,
            'redirect_url': None,
            'tenant_id': self._tenant_id,
        }
        expected.update(extras)
        listener_id = uuidutils.generate_uuid()
        expected['listener_id'] = listener_id

        m.post(self.url, json={'l7policy': expected})
        with self.l7policy(listener_id, no_delete=True) as p:
            actual = {}
            for k, v in p['l7policy'].items():
                if k in expected:
                    actual[k] = v
            self.assertEqual(actual, expected)

    @requests_mock.mock()
    def test_create_l7policy_pool_redirect(self, m, **extras):
        expected = {
            'action': lb_const.L7_POLICY_ACTION_REDIRECT_TO_POOL,
            'redirect_pool_id': None,
            'redirect_url': None,
            'tenant_id': self._tenant_id,
        }
        expected.update(extras)
        listener_id = uuidutils.generate_uuid()
        pool_id = uuidutils.generate_uuid()
        expected['listener_id'] = listener_id
        expected['redirect_pool_id'] = pool_id

        m.post(self.url, json={'l7policy': expected})
        with self.l7policy(
            listener_id,
            action=lb_const.L7_POLICY_ACTION_REDIRECT_TO_POOL,
            redirect_pool_id=pool_id, no_delete=True) as p:
                actual = {}
                for k, v in p['l7policy'].items():
                    if k in expected:
                        actual[k] = v
                self.assertEqual(actual, expected)

    @requests_mock.mock()
    def test_update_l7policy(self, m, **extras):
        l7policy_id = uuidutils.generate_uuid()
        expected = {
            'admin_state_up': False,
            'action': lb_const.L7_POLICY_ACTION_REDIRECT_TO_URL,
            'redirect_pool_id': None,
            'redirect_url': 'redirect_url',
            'tenant_id': self._tenant_id,
            'position': 1,
        }
        expected.update(extras)

        m.get('{}/{}'.format(self.url, l7policy_id),
              json={'l7policy': expected})
        m.put('{}/{}'.format(self.url, l7policy_id),
              json={'l7policy': expected})

        data = {
            'l7policy': {
                'action': lb_const.L7_POLICY_ACTION_REDIRECT_TO_URL,
                'redirect_url': 'redirect_url',
                'admin_state_up': False}}

        ctx = context.get_admin_context()
        self.plugin.update_l7policy(ctx, l7policy_id, data)
        l7policy = self.plugin.get_l7policy(ctx, l7policy_id)
        actual = {}
        for k, v in l7policy.items():
            if k in expected:
                actual[k] = v
        self.assertEqual(actual, expected)

    @requests_mock.mock()
    def test_delete_l7policy(self, m, **extras):
        l7policy_id = uuidutils.generate_uuid()

        m.delete('{}/{}'.format(self.url, l7policy_id))

        c = context.get_admin_context()
        self.plugin.delete_l7policy(c, l7policy_id)

    @requests_mock.mock()
    def test_show_l7policy(self, m, **extras):
        listener_id = uuidutils.generate_uuid()
        l7policy_id = uuidutils.generate_uuid()
        expected = {
            'position': 1,
            'action': lb_const.L7_POLICY_ACTION_REJECT,
            'redirect_pool_id': None,
            'redirect_url': None,
            'tenant_id': self._tenant_id,
            'listener_id': listener_id,
            'id': l7policy_id,
        }
        expected.update(extras)

        m.get('{}/{}'.format(self.url, l7policy_id),
            json={'l7policy': expected})

        req = self.new_show_request('l7policies',
                                    l7policy_id,
                                    fmt=self.fmt)
        res = self.deserialize(self.fmt,
                               req.get_response(self.ext_api))
        actual = {}
        for k, v in res['l7policy'].items():
            if k in expected:
                actual[k] = v
        self.assertEqual(expected, actual)

    @requests_mock.mock()
    def test_list_l7policies_with_sort_emulated(self, m):
        listener_id = uuidutils.generate_uuid()
        l7policy_id = uuidutils.generate_uuid()
        expected = {
            'position': 1,
            'action': lb_const.L7_POLICY_ACTION_REJECT,
            'redirect_pool_id': None,
            'redirect_url': None,
            'tenant_id': self._tenant_id,
            'listener_id': listener_id,
            'id': l7policy_id,
        }
        m.get(self.url, json={'l7policies': [expected]})
        self._test_list_with_sort('l7policy', [{'l7policy': expected}],
                                  [('name', 'asc')],
                                  resources='l7policies')

    @requests_mock.mock()
    def test_create_l7rule_invalid_policy_id(self, m, **extras):
        l7_policy_id = uuidutils.generate_uuid()
        m.post(self._rules(l7_policy_id), status_code=404)
        self._create_l7policy_rule(
            self.fmt, l7_policy_id,
            lb_const.L7_RULE_TYPE_HOST_NAME,
            lb_const.L7_RULE_COMPARE_TYPE_REGEX,
            'value',
            expected_res_status=webob.exc.HTTPNotFound.code)

    @requests_mock.mock()
    def test_create_l7rule(self, m, **extras):
        l7_policy_id = uuidutils.generate_uuid()
        expected = {
            'type': lb_const.L7_RULE_TYPE_HOST_NAME,
            'compare_type': lb_const.L7_RULE_COMPARE_TYPE_EQUAL_TO,
            'key': None,
            'value': 'value1'
        }

        m.post(self._rules(l7_policy_id), json={'rule': expected})
        with self.l7policy_rule(l7_policy_id, no_delete=True) as r_def:
            self.assertEqual(expected, r_def['rule'])

    @requests_mock.mock()
    def test_update_l7rule(self, m, **extras):
        l7_policy_id = uuidutils.generate_uuid()
        l7_policy_rule_id = uuidutils.generate_uuid()

        expected = {}
        expected['type'] = lb_const.L7_RULE_TYPE_HEADER
        expected['compare_type'] = (
            lb_const.L7_RULE_COMPARE_TYPE_REGEX)
        expected['value'] = '/.*/'
        expected['key'] = 'HEADER1'
        expected['invert'] = True
        expected['admin_state_up'] = False

        m.get("{}/{}".format(self._rules(l7_policy_id),
                             l7_policy_rule_id),
              json={'rule': expected})
        m.put("{}/{}".format(self._rules(l7_policy_id),
                             l7_policy_rule_id),
              json={'rule': expected})
        req = self.new_update_request(
            'l7policies', {'rule': expected},
            l7_policy_id, subresource='rules',
            sub_id=l7_policy_rule_id)
        res = self.deserialize(
            self.fmt,
            req.get_response(self.ext_api)
        )
        actual = {}
        for k, v in res['rule'].items():
            if k in expected:
                actual[k] = v
        self.assertEqual(actual, expected)

    @requests_mock.mock()
    def test_delete_l7rule(self, m):
        l7_policy_id = uuidutils.generate_uuid()
        l7_policy_rule_id = uuidutils.generate_uuid()

        m.get("{}/{}".format(
            self._rules(l7_policy_id), l7_policy_rule_id),
            json={'rule': {}})
        m.delete("{}/{}".format(self._rules(l7_policy_id),
                             l7_policy_rule_id))

        req = self.new_delete_request('l7policies',
                                      l7_policy_id,
                                      subresource='rules',
                                      sub_id=l7_policy_rule_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int,
                         webob.exc.HTTPNoContent.code)

    @requests_mock.mock()
    def test_list_l7rules_with_sort_emulated(self, m):
        l7_policy_id = uuidutils.generate_uuid()
        expected = {
            'position': 1,
            'action': lb_const.L7_POLICY_ACTION_REJECT,
            'redirect_pool_id': None,
            'redirect_url': None,
            'tenant_id': self._tenant_id,
            'listener_id': uuidutils.generate_uuid(),
            'id': uuidutils.generate_uuid(),
        }

        m.get(self._rules(l7_policy_id),
              json={'rules': [expected]})

        self._test_list_with_sort('l7policy', [{'rule': expected}],
                                  [('value', 'asc')],
                                  id=l7_policy_id,
                                  resources='l7policies',
                                  subresource='rule',
                                  subresources='rules')


class PoolTestBase(ListenerTestBase):

    def _create_pool_api(self, data):
        req = self.new_create_request("pools", data, self.fmt)
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body

    def _update_pool_api(self, pool_id, data):
        req = self.new_update_request('pools', data, pool_id)
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body

    def _delete_pool_api(self, pool_id):
        req = self.new_delete_request('pools', pool_id)
        resp = req.get_response(self.ext_api)
        return resp

    def _get_pool_api(self, pool_id):
        req = self.new_show_request('pools', pool_id)
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body

    def _list_pools_api(self):
        req = self.new_list_request('pools')
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body


class LbaasPoolTests(PoolTestBase):
    url = '{}/{}'.format(base_url, 'pools')
    listener_id = uuidutils.generate_uuid()

    @requests_mock.mock()
    def test_create_pool(self, m, **extras):
        expected = {
            'name': '',
            'description': '',
            'protocol': 'HTTP',
            'lb_algorithm': 'ROUND_ROBIN',
            'admin_state_up': True,
            'tenant_id': self._tenant_id,
            'healthmonitor_id': None,
            'members': [],
            'id': uuidutils.generate_uuid()
        }

        expected.update(extras)
        m.post(self.url, json={'pool': expected})
        m.get(self.url, json={'pools': [expected]})

        with self.pool(listener_id=self.listener_id, no_delete=True,
                       **extras) as pool:
            pool_id = pool['pool'].get('id')
            if ('session_persistence' in expected.keys() and
                    expected['session_persistence'] is not None and
                    not expected['session_persistence'].get('cookie_name')):
                expected['session_persistence']['cookie_name'] = None
            self.assertTrue(pool_id)

            actual = {}
            for k, v in pool['pool'].items():
                if k in expected:
                    actual[k] = v
            self.assertEqual(expected, actual)
        return pool

    @requests_mock.mock()
    def test_show_pool(self, m, **extras):
        pool_id = uuidutils.generate_uuid()
        expected = {
            'name': '',
            'description': '',
            'protocol': 'HTTP',
            'lb_algorithm': 'ROUND_ROBIN',
            'admin_state_up': True,
            'tenant_id': self._tenant_id,
            'listeners': [{'id': self.listener_id}],
            'healthmonitor_id': None,
            'members': []
        }

        expected.update(extras)

        m.get('{}/{}'.format(self.url, pool_id), json={'pool': expected})
        resp, body = self._get_pool_api(pool_id)
        actual = {}
        for k, v in body['pool'].items():
            if k in expected:
                actual[k] = v
        self.assertEqual(expected, actual)
        return resp

    @requests_mock.mock()
    def test_update_pool(self, m, **extras):
        pool_id = uuidutils.generate_uuid()
        expected = {
            'name': '',
            'description': '',
            'protocol': 'HTTP',
            'lb_algorithm': 'LEAST_CONNECTIONS',
            'admin_state_up': True,
            'tenant_id': self._tenant_id,
            'listeners': [{'id': self.listener_id}],
            'healthmonitor_id': None,
            'members': []
        }

        expected.update(extras)
        m.get('{}/{}'.format(self.url, pool_id), json={'pool': expected})
        m.put('{}/{}'.format(self.url, pool_id), json={'pool': expected})

        data = {'pool': {'lb_algorithm': 'LEAST_CONNECTIONS'}}
        resp, body = self._update_pool_api(pool_id, data)
        actual = {}
        for k, v in body['pool'].items():
            if k in expected:
                actual[k] = v
        self.assertEqual(expected, actual)

        return resp

    @requests_mock.mock()
    def test_delete_pool(self, m):
        pool_id = uuidutils.generate_uuid()
        m.get('{}/{}'.format(self.url, pool_id), json={'pool': {}})
        m.delete('{}/{}'.format(self.url, pool_id))
        resp = self._delete_pool_api(pool_id)
        self.assertEqual(webob.exc.HTTPNoContent.code, resp.status_int)

    @requests_mock.mock()
    def test_cannot_add_multiple_pools_to_listener(self, m):
        data = {'pool': {'name': '',
                         'description': '',
                         'protocol': 'HTTP',
                         'lb_algorithm': 'ROUND_ROBIN',
                         'admin_state_up': True,
                         'tenant_id': self._tenant_id,
                         'listener_id': self.listener_id}}
        m.get(self.url, json={'pools': [data]})
        m.post(self.url, status_code=webob.exc.HTTPConflict.code)
        resp, body = self._create_pool_api(data)
        self.assertEqual(webob.exc.HTTPConflict.code, resp.status_int)

    @requests_mock.mock()
    def test_create_pool_with_protocol_invalid(self, m):
        data = {'pool': {
            'name': '',
            'description': '',
            'protocol': 'BLANK',
            'lb_algorithm': 'LEAST_CONNECTIONS',
            'admin_state_up': True,
            'tenant_id': self._tenant_id
        }}
        m.get(self.url, json={'pools': [data]})
        m.post(self.url, status_code=webob.exc.HTTPBadRequest.code)
        resp, body = self._create_pool_api(data)
        self.assertEqual(webob.exc.HTTPBadRequest.code, resp.status_int)

    @requests_mock.mock()
    def test_list_pools(self, m):
        name = 'list_pools'
        expected_values = {'name': name,
                           'protocol': 'HTTP',
                           'description': 'apool',
                           'lb_algorithm': 'ROUND_ROBIN',
                           'admin_state_up': True,
                           'tenant_id': self._tenant_id,
                           'session_persistence': {'cookie_name': None,
                                                   'type': 'HTTP_COOKIE'},
                           'loadbalancers': [{'id': self.lb_id}],
                           'members': []}

        m.get(self.url, json={'pools': [expected_values]})
        resp, body = self._list_pools_api()
        pool_list = body['pools']
        self.assertEqual(1, len(pool_list))
        for k in expected_values:
            self.assertEqual(expected_values[k], pool_list[0][k])


class MemberTestBase(PoolTestBase):

    def _create_member_api(self, pool_id, data):
        req = self.new_create_request("pools", data, self.fmt, id=pool_id,
                                      subresource='members')
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body

    def _update_member_api(self, pool_id, member_id, data):
        req = self.new_update_request('pools', data, pool_id,
                                      subresource='members', sub_id=member_id)
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body

    def _delete_member_api(self, pool_id, member_id):
        req = self.new_delete_request('pools', pool_id, subresource='members',
                                      sub_id=member_id)
        resp = req.get_response(self.ext_api)
        return resp

    def _get_member_api(self, pool_id, member_id):
        req = self.new_show_request('pools', pool_id, subresource='members',
                                    sub_id=member_id)
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body

    def _list_members_api(self, pool_id):
        req = self.new_list_request('pools', id=pool_id, subresource='members')
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body


class LbaasMemberTests(MemberTestBase):
    pool_id = uuidutils.generate_uuid()
    member_id = uuidutils.generate_uuid()
    url = '{}/{}'.format(base_url, 'pools')
    test_subnet_id = uuidutils.generate_uuid()

    def _members(self, pool_id=None, member_id=None):
        pool_id = pool_id if pool_id else self.pool_id
        member = '/{}'.format(member_id) if member_id else ''
        return "{}/{}/members{}".format(self.url, pool_id, member)

    @requests_mock.mock()
    def test_create_member(self, m, **extras):
        network = self._make_network(self.fmt, 'test-net', True)
        self.test_subnet = self._make_subnet(
            self.fmt, network, gateway=n_constants.ATTR_NOT_SPECIFIED,
            cidr='10.0.0.0/24')
        self.test_subnet_id = self.test_subnet['subnet']['id']
        expected = {
            'address': '127.0.0.1',
            'protocol_port': 80,
            'weight': 1,
            'admin_state_up': True,
            'tenant_id': self._tenant_id,
            'subnet_id': '',
            'name': 'member1',
            'id': uuidutils.generate_uuid()
        }

        expected.update(extras)
        expected['subnet_id'] = self.test_subnet_id

        m.post(self._members(), json={'member': expected})
        with self.member(pool_id=self.pool_id, name='member1',
                         no_delete=True) as member:
            member_id = member['member'].get('id')
            self.assertTrue(member_id)

            actual = {}
            for k, v in member['member'].items():
                if k in expected:
                    actual[k] = v
            self.assertEqual(expected, actual)
        return member

    @requests_mock.mock()
    def test_update_member(self, m):
        expected = {
            'address': '127.0.0.1',
            'protocol_port': 80,
            'weight': 1,
            'admin_state_up': True,
            'tenant_id': self._tenant_id,
            'subnet_id': '',
            'name': 'member1',
            'id': uuidutils.generate_uuid()
        }

        m.get(self._members(member_id=self.member_id),
              json={'member': expected})
        m.put(self._members(member_id=self.member_id),
              json={'member': expected})

        data = {'member': {'weight': 10, 'admin_state_up': False,
                           'name': 'member2'}}
        resp, body = self._update_member_api(self.pool_id,
                                             self.member_id, data)

        actual = {}
        for k, v in body['member'].items():
            if k in expected:
                actual[k] = v
        self.assertEqual(expected, actual)

    @requests_mock.mock()
    def test_delete_member(self, m):
        m.get(self._members(member_id=self.member_id), json={'member': {}})
        m.delete(self._members(member_id=self.member_id))
        resp = self._delete_member_api(self.pool_id, self.member_id)
        self.assertEqual(webob.exc.HTTPNoContent.code, resp.status_int)

    @requests_mock.mock()
    def test_show_member(self, m):
        expected = {
            'address': '127.0.0.1',
            'protocol_port': 80,
            'weight': 1,
            'admin_state_up': True,
            'tenant_id': self._tenant_id,
            'subnet_id': '',
            'name': 'member1',
            'id': uuidutils.generate_uuid()
        }

        m.get(self._members(member_id=self.member_id),
              json={'member': expected})
        resp, body = self._get_member_api(self.pool_id, self.member_id)
        actual = {}
        for k, v in body['member'].items():
            if k in expected:
                actual[k] = v
        self.assertEqual(expected, actual)

    @requests_mock.mock()
    def test_list_members(self, m):
        expected = {
            'address': '127.0.0.1',
            'protocol_port': 80,
            'weight': 1,
            'admin_state_up': True,
            'tenant_id': self._tenant_id,
            'subnet_id': '',
            'name': 'member1',
            'id': uuidutils.generate_uuid()
        }
        m.get(self._members(), json={'members': [expected]})
        resp, body = self._list_members_api(self.pool_id)
        self.assertEqual(1, len(body['members']))

    @requests_mock.mock()
    def test_create_member_invalid_pool_id(self, m):
        data = {'member': {'address': '127.0.0.1',
                           'protocol_port': 80,
                           'weight': 1,
                           'admin_state_up': True,
                           'tenant_id': self._tenant_id,
                           'subnet_id': self.test_subnet_id}}
        m.post(self._members(pool_id='WRONG_POOL_ID'), status_code=404)
        resp, body = self._create_member_api('WRONG_POOL_ID', data)
        self.assertEqual(webob.exc.HTTPNotFound.code, resp.status_int)

    @requests_mock.mock()
    def test_create_member_invalid_name(self, m):
        m.post(self._members(), status_code=webob.exc.HTTPBadRequest.code)
        data = {'member': {'address': '127.0.0.1',
                           'protocol_port': 80,
                           'weight': 1,
                           'admin_state_up': True,
                           'tenant_id': self._tenant_id,
                           'subnet_id': self.test_subnet_id,
                           'name': 123}}
        resp, body = self._create_member_api('POOL_ID', data)
        self.assertEqual(webob.exc.HTTPBadRequest.code, resp.status_int)

    @requests_mock.mock()
    def test_delete_member_invalid_pool_id(self, m):
        m.get(self._members('WRONG_POOL_ID', self.member_id), status_code=404)

        resp = self._delete_member_api('WRONG_POOL_ID', self.member_id)
        self.assertEqual(webob.exc.HTTPNotFound.code, resp.status_int)


class HealthMonitorTestBase(MemberTestBase):

    def _create_healthmonitor_api(self, data):
        req = self.new_create_request("healthmonitors", data, self.fmt)
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body

    def _update_healthmonitor_api(self, hm_id, data):
        req = self.new_update_request('healthmonitors', data, hm_id)
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body

    def _delete_healthmonitor_api(self, hm_id):
        req = self.new_delete_request('healthmonitors', hm_id)
        resp = req.get_response(self.ext_api)
        return resp

    def _get_healthmonitor_api(self, hm_id):
        req = self.new_show_request('healthmonitors', hm_id)
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body

    def _list_healthmonitors_api(self):
        req = self.new_list_request('healthmonitors')
        resp = req.get_response(self.ext_api)
        body = self.deserialize(self.fmt, resp)
        return resp, body


class TestLbaasHealthMonitorTests(HealthMonitorTestBase):
    url = '{}/{}'.format(base_url, 'healthmonitors')
    hm_id = uuidutils.generate_uuid()
    pool_id = uuidutils.generate_uuid()

    @requests_mock.mock()
    def test_create_healthmonitor(self, m, **extras):
        expected = {
            'type': 'HTTP',
            'delay': 1,
            'timeout': 1,
            'max_retries': 2,
            'http_method': 'GET',
            'url_path': '/',
            'expected_codes': '200',
            'admin_state_up': True,
            'tenant_id': self._tenant_id,
            'pools': [{'id': self.pool_id}],
            'name': 'monitor1',
            'id': self.hm_id
        }

        expected.update(extras)
        m.post(self.url, json={'healthmonitor': expected})

        with self.healthmonitor(pool_id=self.pool_id, type='HTTP',
                                name='monitor1', no_delete=True,
                                **extras) as healthmonitor:
            hm_id = healthmonitor['healthmonitor'].get('id')
            self.assertTrue(hm_id)

            actual = {}
            for k, v in healthmonitor['healthmonitor'].items():
                if k in expected:
                    actual[k] = v
            self.assertEqual(expected, actual)
        return healthmonitor

    @requests_mock.mock()
    def test_show_healthmonitor(self, m, **extras):
        expected = {
            'type': 'HTTP',
            'delay': 1,
            'timeout': 1,
            'max_retries': 2,
            'http_method': 'GET',
            'url_path': '/',
            'expected_codes': '200',
            'admin_state_up': True,
            'tenant_id': self._tenant_id,
            'pools': [{'id': self.pool_id}],
            'name': 'monitor1'
        }

        expected.update(extras)

        m.get('{}/{}'.format(self.url, self.hm_id),
              json={'healthmonitor': expected})

        resp, body = self._get_healthmonitor_api(self.hm_id)
        actual = {}
        for k, v in body['healthmonitor'].items():
            if k in expected:
                actual[k] = v
        self.assertEqual(expected, actual)

    @requests_mock.mock()
    def test_update_healthmonitor(self, m, **extras):
        expected = {
            'type': 'HTTP',
            'delay': 30,
            'timeout': 10,
            'max_retries': 4,
            'http_method': 'GET',
            'url_path': '/index.html',
            'expected_codes': '200,404',
            'admin_state_up': True,
            'tenant_id': self._tenant_id,
            'pools': [{'id': self.pool_id}],
            'name': 'monitor2'
        }

        expected.update(extras)
        m.get('{}/{}'.format(self.url, self.hm_id),
              json={'healthmonitor': expected})
        m.put('{}/{}'.format(self.url, self.hm_id),
              json={'healthmonitor': expected})

        data = {'healthmonitor': {'delay': 30,
                                  'timeout': 10,
                                  'max_retries': 4,
                                  'expected_codes': '200,404',
                                  'url_path': '/index.html',
                                  'name': 'monitor2'}}
        resp, body = self._update_healthmonitor_api(self.hm_id, data)
        actual = {}
        for k, v in body['healthmonitor'].items():
            if k in expected:
                actual[k] = v
        self.assertEqual(expected, actual)

    @requests_mock.mock()
    def test_delete_healthmonitor(self, m):
        m.get('{}/{}'.format(self.url, self.hm_id), json={'healthmonitor': {}})
        m.delete('{}/{}'.format(self.url, self.hm_id))
        resp = self._delete_healthmonitor_api(self.hm_id)
        self.assertEqual(webob.exc.HTTPNoContent.code, resp.status_int)

    @requests_mock.mock()
    def test_create_health_monitor_with_timeout_invalid(self, m):
        m.post(self.url, status_code=webob.exc.HTTPBadRequest.code)
        data = {'healthmonitor': {'type': 'HTTP',
                                  'delay': 1,
                                  'timeout': -1,
                                  'max_retries': 2,
                                  'admin_state_up': True,
                                  'tenant_id': self._tenant_id,
                                  'pool_id': self.pool_id}}
        resp, body = self._create_healthmonitor_api(data)
        self.assertEqual(webob.exc.HTTPBadRequest.code, resp.status_int)

    @requests_mock.mock()
    def test_create_healthmonitor_invalid_pool_id(self, m):
        m.post(self.url, status_code=webob.exc.HTTPNotFound.code)
        data = {'healthmonitor': {'type': lb_const.HEALTH_MONITOR_TCP,
                                  'delay': 1,
                                  'timeout': 1,
                                  'max_retries': 1,
                                  'tenant_id': self._tenant_id,
                                  'pool_id': uuidutils.generate_uuid()}}
        resp, body = self._create_healthmonitor_api(data)
        self.assertEqual(webob.exc.HTTPNotFound.code, resp.status_int)

    @requests_mock.mock()
    def test_only_one_healthmonitor_per_pool(self, m):
        m.post(self.url, status_code=webob.exc.HTTPConflict.code)
        data = {'healthmonitor': {'type': lb_const.HEALTH_MONITOR_TCP,
                                  'delay': 1,
                                  'timeout': 1,
                                  'max_retries': 1,
                                  'tenant_id': self._tenant_id,
                                  'pool_id': self.pool_id}}
        resp, body = self._create_healthmonitor_api(data)
        self.assertEqual(webob.exc.HTTPConflict.code, resp.status_int)

    @requests_mock.mock()
    def test_list_healthmonitors(self, m):
        expected = {
            'type': 'HTTP',
            'delay': 1,
            'timeout': 1,
            'max_retries': 2,
            'http_method': 'GET',
            'url_path': '/',
            'expected_codes': '200',
            'admin_state_up': True,
            'tenant_id': self._tenant_id,
            'pools': [{'id': self.pool_id}],
            'name': '',
            'max_retries_down': 3,
            'id': self.hm_id
        }

        m.get(self.url, json={'healthmonitors': [expected]})

        resp, body = self._list_healthmonitors_api()
        self.assertEqual([expected], body['healthmonitors'])


class LbaasStatusesTest(TestLbaasProxyPluginDbTestCase):
    lb_id = uuidutils.generate_uuid()
    url = '{}/{}'.format(base_url, 'loadbalancers')

    @requests_mock.mock()
    def test_status_tree_lb(self, m):
        expected = {
            "loadbalancer": {
                "operating_status": "ONLINE",
                "provisioning_status": "ACTIVE",
                "listeners": [
                    {
                        "id": "6978ba19-1090-48a2-93d5-3523592b562a",
                        "operating_status": "ONLINE",
                        "provisioning_status": "ACTIVE",
                        "pools": [
                            {
                                "id": "f6aedfcb-9f7d-4cc5-83e1-8c02fd833922",
                                "operating_status": "ONLINE",
                                "provisioning_status": "ACTIVE",
                                "members": [
                                    {
                                        "id":
                                       "fcf23bde-8cf9-4616-883f-208cebcbf858",
                                        "operating_status": "ONLINE",
                                        "provisioning_status": "ACTIVE",
                                    }
                                ],
                                "healthmonitor": {
                                    "id":
                                        "785131d2-8f7b-4fee-a7e7-3196e11b4518",
                                    "provisioning_status": "ACTIVE",
                                }
                            }
                        ]
                    }
                ]
            }
        }
        m.get('{}/{}'.format(self.url, self.lb_id),
              json={'loadbalancer': {'id': self.lb_id}})
        m.get('{}/{}/status'.format(self.url, self.lb_id),
              json={'statuses': expected})
        statuses = self._get_loadbalancer_statuses_api(self.lb_id)[1]
        self.assertEqual(expected, statuses['statuses'])


class LbaasStatsTest(MemberTestBase):
    lb_id = uuidutils.generate_uuid()
    url = '{}/{}'.format(base_url, 'loadbalancers')

    @requests_mock.mock()
    def test_stats(self, m):
        expected = {
            "stats": {
                "bytes_in": "131342840",
                "total_connections": "52378345",
                "active_connections": "97258",
                "bytes_out": "1549542372",
                "request_errors": "0"
            }
        }
        m.get('{}/{}'.format(self.url, self.lb_id),
              json={'loadbalancer': {'id': self.lb_id}})
        m.get('{}/{}/stats'.format(self.url, self.lb_id),
              json={'stats': expected})
        stats = self._get_loadbalancer_stats_api(self.lb_id)[1]
        self.assertEqual(expected, stats['stats'])


class LbaasGraphTest(MemberTestBase):
    lb_id = uuidutils.generate_uuid()
    url = '{}/{}'.format(base_url, 'graph')

    @requests_mock.mock()
    def create_graph(self, m, expected_lb_graph, listeners):
        with self.subnet() as subnet:
            expected_lb_graph['vip_subnet_id'] = subnet['subnet']['id']
            for listener in listeners:
                for member in listener.get('default_pool',
                                           {}).get('members', []):
                    member['subnet_id'] = subnet['subnet']['id']
            for listener in expected_lb_graph.get('listeners', []):
                for member in listener.get('default_pool',
                                           {}).get('members', []):
                    member['subnet_id'] = subnet['subnet']['id']
            name = expected_lb_graph.get('name')
            kwargs = {'name': name, 'subnet': subnet, 'listeners': listeners}
            m.post(self.url, json={'graph': expected_lb_graph})
            with self.graph(no_delete=True, **kwargs) as graph:
                lb = graph['graph']['loadbalancer']
                self._assert_graphs_equal(expected_lb_graph, lb)



# Copyright 2016 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Add healthmonitor max retries down

Revision ID: 844352f9fe6f
Revises: 62deca5010cd
Create Date: 2016-04-21 15:32:05.647920

"""

from alembic import op
import sqlalchemy as sa

from neutron.db import migration


# revision identifiers, used by Alembic.
revision = '844352f9fe6f'
down_revision = '62deca5010cd'


# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.NEWTON]


def upgrade():
    op.add_column('lbaas_healthmonitors', sa.Column(
        u'max_retries_down', sa.Integer(), nullable=True))



# Copyright 2014 Rackspace US, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os

import mock
from oslo_config import cfg
from oslo_config import fixture as oslo_fixture

from neutron_lbaas.common.cert_manager import cert_manager
from neutron_lbaas.common.cert_manager import local_cert_manager
from neutron_lbaas.tests import base


class TestLocalCert(base.BaseTestCase):

    def setUp(self):
        self.certificate = "My Certificate"
        self.intermediates = "My Intermediates"
        self.private_key = "My Private Key"
        self.private_key_passphrase = "My Private Key Passphrase"

        super(TestLocalCert, self).setUp()

    def test_local_cert(self):
        # Create a cert
        cert = local_cert_manager.Cert(
            certificate=self.certificate,
            intermediates=self.intermediates,
            private_key=self.private_key,
            private_key_passphrase=self.private_key_passphrase
        )

        # Validate the cert functions
        self.assertEqual(self.certificate, cert.get_certificate())
        self.assertEqual(self.intermediates, cert.get_intermediates())
        self.assertEqual(self.private_key, cert.get_private_key())
        self.assertEqual(self.private_key_passphrase,
                         cert.get_private_key_passphrase())


class TestLocalManager(base.BaseTestCase):

    def setUp(self):
        self.project_id = "12345"
        self.certificate = "My Certificate"
        self.intermediates = "My Intermediates"
        self.private_key = "My Private Key"
        self.private_key_passphrase = "My Private Key Passphrase"

        conf = oslo_fixture.Config(cfg.CONF)
        conf.config(group="certificates", storage_path="/tmp/")

        self.cert_manager = local_cert_manager.CertManager()

        super(TestLocalManager, self).setUp()

    def _store_cert(self):
        file_mock = mock.mock_open()
        # Attempt to store the cert
        with mock.patch('six.moves.builtins.open', file_mock, create=True):
            cert_id = self.cert_manager.store_cert(
                project_id=self.project_id,
                certificate=self.certificate,
                intermediates=self.intermediates,
                private_key=self.private_key,
                private_key_passphrase=self.private_key_passphrase
            )

        # Check that something came back
        self.assertIsNotNone(cert_id)

        # Verify the correct files were opened
        file_mock.assert_has_calls([
            mock.call(os.path.join('/tmp/{0}.crt'.format(cert_id)), 'w'),
            mock.call(os.path.join('/tmp/{0}.key'.format(cert_id)), 'w'),
            mock.call(os.path.join('/tmp/{0}.int'.format(cert_id)), 'w'),
            mock.call(os.path.join('/tmp/{0}.pass'.format(cert_id)), 'w')
        ], any_order=True)

        # Verify the writes were made
        file_mock().write.assert_has_calls([
            mock.call(self.certificate),
            mock.call(self.intermediates),
            mock.call(self.private_key),
            mock.call(self.private_key_passphrase)
        ], any_order=True)

        return cert_id

    def _get_cert(self, cert_id):
        file_mock = mock.mock_open()
        # Attempt to retrieve the cert
        with mock.patch('six.moves.builtins.open', file_mock, create=True):
            data = self.cert_manager.get_cert(
                project_id=self.project_id,
                cert_ref=cert_id,
                resource_ref=None
            )

        # Verify the correct files were opened
        file_mock.assert_has_calls([
            mock.call(os.path.join('/tmp/{0}.crt'.format(cert_id)), 'r'),
            mock.call(os.path.join('/tmp/{0}.key'.format(cert_id)), 'r'),
            mock.call(os.path.join('/tmp/{0}.int'.format(cert_id)), 'r'),
            mock.call(os.path.join('/tmp/{0}.pass'.format(cert_id)), 'r')
        ], any_order=True)

        # The returned data should be a Cert object
        self.assertIsInstance(data, cert_manager.Cert)

        return data

    def _delete_cert(self, cert_id):
        remove_mock = mock.Mock()
        # Delete the cert
        with mock.patch('os.remove', remove_mock):
            self.cert_manager.delete_cert(
                project_id=self.project_id,
                cert_ref=cert_id,
                resource_ref=None
            )

        # Verify the correct files were removed
        remove_mock.assert_has_calls([
            mock.call(os.path.join('/tmp/{0}.crt'.format(cert_id))),
            mock.call(os.path.join('/tmp/{0}.key'.format(cert_id))),
            mock.call(os.path.join('/tmp/{0}.int'.format(cert_id))),
            mock.call(os.path.join('/tmp/{0}.pass'.format(cert_id)))
        ], any_order=True)

    def test_store_cert(self):
        self._store_cert()

    def test_get_cert(self):
        # Store a cert
        cert_id = self._store_cert()

        # Get the cert
        self._get_cert(cert_id)

    def test_delete_cert(self):
        # Store a cert
        cert_id = self._store_cert()

        # Verify the cert exists
        self._get_cert(cert_id)

        # Delete the cert
        self._delete_cert(cert_id)



# Copyright 2014-2016 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron_lbaas.common.cert_manager.barbican_auth import barbican_acl
from neutron_lbaas.common.cert_manager import barbican_cert_manager
from neutron_lbaas.common import keystone
import neutron_lbaas.tests.base as base


class TestBarbicanACLAuth(base.BaseTestCase):
    def setUp(self):
        # Reset the client
        keystone._SESSION = None

        super(TestBarbicanACLAuth, self).setUp()

    def test_get_barbican_client(self):
        # There should be no existing client
        self.assertIsNone(keystone._SESSION)

        # Mock out the keystone session and get the client
        keystone._SESSION = mock.MagicMock()
        acl_auth_object = barbican_acl.BarbicanACLAuth()
        bc1 = acl_auth_object.get_barbican_client()

        # Our returned object should have elements that proves it is a real
        # Barbican client object. We shouldn't use `isinstance` because that's
        # an evil pattern, instead we can check for very unique things in the
        # stable client API like "register_consumer", since this should fairly
        # reliably prove we're dealing with a Barbican client.
        self.assertTrue(hasattr(bc1, 'containers') and
                        hasattr(bc1.containers, 'register_consumer'))

        # Getting the session again with new class should get the same object
        acl_auth_object2 = barbican_acl.BarbicanACLAuth()
        bc2 = acl_auth_object2.get_barbican_client()
        self.assertIs(bc1, bc2)

    def test_load_auth_driver(self):
        bcm = barbican_cert_manager.CertManager()
        self.assertIsInstance(bcm.auth, barbican_acl.BarbicanACLAuth)



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron_lib import context

from neutron_lbaas.drivers.logging_noop import driver
from neutron_lbaas.services.loadbalancer import data_models
from neutron_lbaas.tests.unit.db.loadbalancer import test_db_loadbalancerv2

log_path = ('neutron_lbaas.drivers.logging_noop.driver.LOG')


class FakeModel(object):
    def __init__(self, id):
        self.id = id

    def attached_to_loadbalancer(self):
        return True


def patch_manager(func):
    @mock.patch(log_path)
    def wrapper(*args):
        log_mock = args[-1]
        manager_test = args[0]
        model = args[1]
        parent = manager_test.parent
        driver = parent.driver
        driver.plugin.reset_mock()

        func(*args[:-1])

        s = str(log_mock.mock_calls[0])
        parent.assertEqual("call.debug(", s[:11])
        parent.assertTrue(s.index(model.id) != -1,
                          msg="Model ID not found in log")

    return wrapper


class ManagerTest(object):
    def __init__(self, parent, manager, model):
        self.parent = parent
        self.manager = manager

        self.create(model)
        self.update(model, model)
        self.delete(model)

    @patch_manager
    def create(self, model):
        self.manager.create(self.parent.context, model)

    @patch_manager
    def update(self, old_model, model):
        self.manager.update(self.parent.context, old_model, model)

    @patch_manager
    def delete(self, model):
        self.manager.delete(self.parent.context, model)


class ManagerTestWithUpdates(ManagerTest):
    def __init__(self, parent, manager, model):
        self.parent = parent
        self.manager = manager

        self.create(model)
        self.update(model, model)
        self.delete(model)

    @patch_manager
    def create(self, model):
        self.manager.create(self.parent.context, model)

    @patch_manager
    def update(self, old_model, model):
        self.manager.update(self.parent.context, old_model, model)

    @patch_manager
    def delete(self, model):
        self.manager.delete(self.parent.context, model)


class LoadBalancerManagerTest(ManagerTestWithUpdates):
    def __init__(self, parent, manager, model):
        super(LoadBalancerManagerTest, self).__init__(parent, manager, model)

        self.create_and_allocate_vip(model)
        self.refresh(model)
        self.stats(model)

    @patch_manager
    def allocates_vip(self):
        self.manager.allocates_vip()

    @patch_manager
    def create_and_allocate_vip(self, model):
        self.manager.create(self.parent.context, model)

    @patch_manager
    def refresh(self, model):
        self.manager.refresh(self.parent.context, model)

    @patch_manager
    def stats(self, model):
        dummy_stats = {
            "bytes_in": 0,
            "bytes_out": 0,
            "active_connections": 0,
            "total_connections": 0
        }
        h = self.manager.stats(self.parent.context, model)
        self.parent.assertEqual(dummy_stats, h)


class TestLoggingNoopLoadBalancerDriver(
        test_db_loadbalancerv2.LbaasPluginDbTestCase):

    def _create_fake_models(self):
        id = 'name-001'
        lb = data_models.LoadBalancer(id=id)
        pool = data_models.Pool(id=id, loadbalancer=lb)
        listener = data_models.Listener(id=id, loadbalancer=lb)
        member = data_models.Member(id=id, pool=pool)
        hm = data_models.HealthMonitor(id=id, pool=pool)
        lb.listeners = [listener]
        lb.pools = [pool]
        listener.default_pool = pool
        pool.members = [member]
        pool.healthmonitor = hm
        return lb

    def setUp(self):
        super(TestLoggingNoopLoadBalancerDriver, self).setUp()
        self.context = context.get_admin_context()
        self.plugin = mock.Mock()
        self.driver = driver.LoggingNoopLoadBalancerDriver(self.plugin)
        self.lb = self._create_fake_models()

    def test_load_balancer_ops(self):
        LoadBalancerManagerTest(self, self.driver.load_balancer, self.lb)

    def test_listener_ops(self):
        ManagerTest(self, self.driver.listener, self.lb.listeners[0])

    def test_pool_ops(self):
        ManagerTestWithUpdates(self, self.driver.pool,
                               self.lb.listeners[0].default_pool)

    def test_member_ops(self):
        ManagerTestWithUpdates(self, self.driver.member,
                               self.lb.listeners[0].default_pool.members[0])

    def test_health_monitor_ops(self):
        ManagerTest(self, self.driver.health_monitor,
                    self.lb.listeners[0].default_pool.healthmonitor)



# Copyright 2015 Citrix Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron_lbaas.drivers.netscaler import ncc_client
from neutron_lbaas.drivers.netscaler import netscaler_driver_v2
from neutron_lbaas.services.loadbalancer import data_models
from neutron_lbaas.tests.unit.db.loadbalancer import test_db_loadbalancerv2

LBAAS_DRIVER_CLASS = ('neutron_lbaas.drivers.netscaler.netscaler_driver_v2'
                      '.NetScalerLoadBalancerDriverV2')

NCC_CLIENT_CLASS = ('neutron_lbaas.drivers.netscaler.ncc_client.NSClient')

LBAAS_PROVIDER_NAME = 'NetScaler'
LBAAS_PROVIDER = ('LOADBALANCERV2:%s:%s:default' %
                  (LBAAS_PROVIDER_NAME, LBAAS_DRIVER_CLASS))

log_path = ('neutron_lbaas.drivers.logging_noop.driver.LOG')


class FakeModel(object):

    def __init__(self, id):
        self.id = id

    def attached_to_loadbalancer(self):
        return True


class ManagerTest(object):

    def __init__(self, parent, manager, model):
        self.parent = parent
        self.manager = manager
        self.model = model
        self.object_path = None
        self.async_obj_track_list = (netscaler_driver_v2.
                                     PROVISIONING_STATUS_TRACKER)
        self.successful_completion_mock = mock.patch.object(
            manager, "successful_completion").start()
        self.failed_completion_mock = mock.patch.object(
            manager, "failed_completion").start()

    def start_tests(self):
        model = self.model
        self.object_path = "%s/%s" % (
            self.resource_path,
            model.id)
        self.create_success(model)
        self.update_success(model, model)
        self.delete_success(model)

        self.create_failure(model)
        self.update_failure(model, model)
        self.delete_failure(model)

    def _check_success_completion(self):
        """Check if success_completion is called"""
        successful_completion_mock = self.successful_completion_mock
        successful_completion_mock.assert_called_once_with(
            mock.ANY, self.model)
        successful_completion_mock.reset_mock()

    def _check_success_completion_with_delete(self):
        """Check if success_completion is called with delete"""
        successful_completion_mock = self.successful_completion_mock
        successful_completion_mock.assert_called_once_with(
            mock.ANY, self.model, delete=True)
        successful_completion_mock.reset_mock()

    def _check_failure_completion(self):
        """Check failed_completion is called"""
        failed_completion_mock = self.failed_completion_mock
        failed_completion_mock.assert_called_once_with(
            mock.ANY, self.model)
        failed_completion_mock.reset_mock()

    def _set_response_error(self, mock_instance):
        errorcode = ncc_client.NCCException.RESPONSE_ERROR
        mock_instance.side_effect = (ncc_client
                                     .NCCException(errorcode))

    def create(self, model):
        self.manager.create(self.parent.context, model)
        create_resource_mock = self.parent.create_resource_mock
        self.parent.assertTrue(create_resource_mock.called)
        resource_path = self.resource_path
        object_name = self.object_name
        create_payload = mock.ANY
        create_resource_mock.assert_called_once_with(
            mock.ANY, resource_path, object_name, create_payload)

    def update(self, old_model, model):
        self.manager.update(self.parent.context, old_model, model)
        update_resource_mock = self.parent.update_resource_mock
        self.parent.assertTrue(update_resource_mock.called)
        object_path = self.object_path
        object_name = self.object_name
        update_payload = mock.ANY
        update_resource_mock.assert_called_once_with(
            mock.ANY, object_path, object_name, update_payload)

    def delete(self, model):
        self.manager.delete(self.parent.context, model)
        remove_resource_mock = self.parent.remove_resource_mock
        self.parent.assertTrue(remove_resource_mock.called)
        object_path = self.object_path
        remove_resource_mock.assert_called_once_with(
            mock.ANY, object_path)

    def check_op_status(self, model, delete=False):
        loadbalancer = model.root_loadbalancer
        if hasattr(self, "async_obj_track_list") and self.async_obj_track_list:
            self.parent.assertIn(
                loadbalancer.id, self.async_obj_track_list)
        else:
            if delete:
                self._check_success_completion_with_delete()
            else:
                self._check_success_completion()

    def create_success(self, model):
        self.create(model)
        self.check_op_status(model)
        self.parent.create_resource_mock.reset()

    def update_success(self, old_model, model):
        self.update(old_model, model)
        self.check_op_status(model)
        self.parent.update_resource_mock.reset_mock()

    def delete_success(self, model):
        self.delete(model)
        self.check_op_status(model, delete=True)
        self.parent.remove_resource_mock.reset_mock()

    def create_failure(self, model):
        create_resource_mock = self.parent.create_resource_mock
        self._set_response_error(create_resource_mock)

        try:
            self.create(model)
        except Exception:
            pass

        self._check_failure_completion()
        create_resource_mock.reset_mock()
        create_resource_mock.side_effect = mock_create_resource_func

    def update_failure(self, old_model, model):
        update_resource_mock = self.parent.update_resource_mock
        self._set_response_error(update_resource_mock)

        try:
            self.update(old_model, model)
        except Exception:
            pass

        self._check_failure_completion()
        update_resource_mock.reset_mock()
        update_resource_mock.side_effect = mock_update_resource_func

    def delete_failure(self, model):
        remove_resource_mock = self.parent.remove_resource_mock
        self._set_response_error(remove_resource_mock)
        try:
            self.delete(model)
        except Exception:
            pass

        self._check_failure_completion()
        remove_resource_mock.reset_mock()
        remove_resource_mock.side_effect = mock_remove_resource_func


class LoadBalancerManagerTest(ManagerTest):

    def __init__(self, parent, manager, model):
        super(LoadBalancerManagerTest, self).__init__(parent, manager, model)
        self.object_name = netscaler_driver_v2.LB_RESOURCE
        self.resource_path = "%s/%s" % (
            netscaler_driver_v2.RESOURCE_PREFIX,
            netscaler_driver_v2.LBS_RESOURCE)
        self.start_tests()


class ListenerManagerTest(ManagerTest):

    def __init__(self, parent, manager, model):
        super(ListenerManagerTest, self).__init__(parent, manager, model)
        self.object_name = netscaler_driver_v2.LISTENER_RESOURCE
        self.resource_path = "%s/%s" % (
            netscaler_driver_v2.RESOURCE_PREFIX,
            netscaler_driver_v2.LISTENERS_RESOURCE)
        self.start_tests()


class PoolManagerTest(ManagerTest):

    def __init__(self, parent, manager, model):
        super(PoolManagerTest, self).__init__(parent, manager, model)
        self.object_name = netscaler_driver_v2.POOL_RESOURCE
        self.resource_path = "%s/%s" % (
            netscaler_driver_v2.RESOURCE_PREFIX,
            netscaler_driver_v2.POOLS_RESOURCE)
        self.start_tests()


class MemberManagerTest(ManagerTest):

    def __init__(self, parent, manager, model):
        super(MemberManagerTest, self).__init__(parent, manager, model)
        self.object_name = netscaler_driver_v2.MEMBER_RESOURCE
        self.resource_path = "%s/%s/%s/%s" % (
            netscaler_driver_v2.RESOURCE_PREFIX,
            netscaler_driver_v2.POOLS_RESOURCE,
            model.pool.id,
            netscaler_driver_v2.MEMBERS_RESOURCE)
        self.start_tests()


class MonitorManagerTest(ManagerTest):

    def __init__(self, parent, manager, model):
        super(MonitorManagerTest, self).__init__(parent, manager, model)
        self.object_name = netscaler_driver_v2.MONITOR_RESOURCE
        self.resource_path = "%s/%s" % (
            netscaler_driver_v2.RESOURCE_PREFIX,
            netscaler_driver_v2.MONITORS_RESOURCE)
        self.start_tests()


class TestNetScalerLoadBalancerDriverV2(
        test_db_loadbalancerv2.LbaasPluginDbTestCase):

    def _create_fake_models(self):
        id = 'name-001'
        lb = data_models.LoadBalancer(id=id)
        listener = data_models.Listener(id=id, loadbalancer=lb)
        pool = data_models.Pool(id=id, listener=listener)
        member = data_models.Member(id=id, pool=pool)
        hm = data_models.HealthMonitor(id=id, pool=pool)
        lb.listeners = [listener]
        listener.default_pool = pool
        pool.members = [member]
        pool.healthmonitor = hm
        return lb

    def _get_fake_network_info(self):
        network_info = {}
        network_info["network_id"] = "network_id_1"
        network_info["subnet_id"] = "subnet_id_1"
        return network_info

    def setUp(self):
        super(TestNetScalerLoadBalancerDriverV2, self).setUp()
        self.context = mock.Mock()

        self.plugin = mock.Mock()
        self.lb = self._create_fake_models()
        mock.patch.object(netscaler_driver_v2, 'LOG').start()

        network_info_mock = mock.patch.object(
            netscaler_driver_v2.PayloadPreparer, "get_network_info").start()
        network_info_mock.return_value = self._get_fake_network_info()

        mock.patch.object(
            netscaler_driver_v2.NetScalerLoadBalancerDriverV2,
            "_init_status_collection").start()

        """mock the NSClient class (REST client)"""
        client_mock_cls = mock.patch(NCC_CLIENT_CLASS).start()

        """mock the REST methods of the NSClient class"""

        self.client_mock_instance = client_mock_cls.return_value
        self.create_resource_mock = self.client_mock_instance.create_resource
        self.create_resource_mock.side_effect = mock_create_resource_func
        self.update_resource_mock = self.client_mock_instance.update_resource
        self.update_resource_mock.side_effect = mock_update_resource_func
        self.retrieve_resource_mock = (self.client_mock_instance
                                           .retrieve_resource)
        self.retrieve_resource_mock.side_effect = mock_retrieve_resource_func
        self.remove_resource_mock = self.client_mock_instance.remove_resource
        self.remove_resource_mock.side_effect = mock_remove_resource_func

        self.driver = netscaler_driver_v2.NetScalerLoadBalancerDriverV2(
            self.plugin)
        self.assertTrue(client_mock_cls.called)

    def test_load_balancer_ops(self):
        LoadBalancerManagerTest(self, self.driver.load_balancer, self.lb)

    def test_listener_ops(self):
        ListenerManagerTest(self, self.driver.listener, self.lb.listeners[0])

    def test_pool_ops(self):
        PoolManagerTest(self, self.driver.pool,
                        self.lb.listeners[0].default_pool)

    def test_member_ops(self):
        MemberManagerTest(self, self.driver.member,
                          self.lb.listeners[0].default_pool.members[0])

    def test_health_monitor_ops(self):
        MonitorManagerTest(self, self.driver.health_monitor,
                           self.lb.listeners[0].default_pool.healthmonitor)


def mock_create_resource_func(*args, **kwargs):
    return 201, {}


def mock_update_resource_func(*args, **kwargs):
    return 202, {}


def mock_retrieve_resource_func(*args, **kwargs):
    return 200, {}


def mock_remove_resource_func(*args, **kwargs):
    return 200, {}



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import utils

from neutron_lbaas.tests.tempest.v2.scenario import base


class TestHealthMonitorBasic(base.BaseTestCase):

    @utils.services('compute', 'network')
    def test_health_monitor_basic(self):
        """This test checks load balancing with health monitor.

        The following is the scenario outline:
        1. Create two instances.
        2. SSH to the instances and start two servers: primary and secondary.
        3. Create a load balancer, with two members and with
           ROUND_ROBIN algorithm, associate the VIP with a floating ip.
        4. Create a health monitor.
        5. Send NUM requests to the floating ip and check that they are shared
           between the two servers.
        6. Disable the primary server and validate the traffic is being sent
           only to the secondary server.
        """
        self._create_servers()
        self._start_servers()
        self._create_load_balancer()
        self._create_health_monitor()
        self._check_load_balancing()
        # stopping the primary server
        self._stop_server()
        # Asserting the traffic is sent only to the secondary server
        self._traffic_validation_after_stopping_server()



# Copyright 2016 Rackspace Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import utils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as ex

from neutron_lbaas.tests.tempest.v2.api import base

PROTOCOL_PORT = 80


class TestPools(base.BaseTestCase):

    """
    Tests the following operations in the Neutron-LBaaS API using the
    REST client for Pools:

        list pools
        create pool
        get pool
        update pool
        delete pool
    """

    @classmethod
    def resource_setup(cls):
        super(TestPools, cls).resource_setup()
        if not utils.is_extension_enabled('lbaasv2', 'network'):
            msg = "lbaas extension not enabled."
            raise cls.skipException(msg)
        network_name = data_utils.rand_name('network-')
        cls.network = cls.create_network(network_name)
        cls.subnet = cls.create_subnet(cls.network)
        cls.load_balancer = cls._create_load_balancer(
            tenant_id=cls.subnet.get('tenant_id'),
            vip_subnet_id=cls.subnet.get('id'),
            wait=True)
        cls.listener = cls._create_listener(
            loadbalancer_id=cls.load_balancer.get('id'),
            protocol=cls.listener_protocol, protocol_port=80)

    def increment_protocol_port(self):
        global PROTOCOL_PORT
        PROTOCOL_PORT += 1

    def _prepare_and_create_pool(self, protocol=None, lb_algorithm=None,
                                 listener_id=None, cleanup=True, **kwargs):
        self._wait_for_load_balancer_status(self.load_balancer.get('id'))
        self.increment_protocol_port()
        if not protocol:
            protocol = self.pool_protocol
        if not lb_algorithm:
            lb_algorithm = 'ROUND_ROBIN'
        if not listener_id:
            listener = self._create_listener(
                loadbalancer_id=self.load_balancer.get('id'),
                protocol=self.listener_protocol, protocol_port=PROTOCOL_PORT,
                wait=True)
            listener_id = listener.get('id')
        response = self._create_pool(protocol=protocol,
                                     lb_algorithm=lb_algorithm,
                                     listener_id=listener_id,
                                     wait=True,
                                     **kwargs)
        if cleanup:
            self.addCleanup(self._delete_pool, response['id'])
        return response

    def _prep_list_comparison(self, single, obj_list):
        for obj in obj_list:
            if not single.get('updated_at') and obj.get('updated_at'):
                obj['updated_at'] = None
            self._test_provisioning_status_if_exists(single, obj)

    def test_list_pools_empty(self):
        """Test get pools when empty"""
        pools = self.pools_client.list_pools()
        self.assertEmpty(pools)

    def test_list_pools_one(self):
        """Test get pools with one pool"""
        new_pool = self._prepare_and_create_pool()
        pools = self.pools_client.list_pools()
        self.assertEqual(1, len(pools))
        self._prep_list_comparison(new_pool, pools)
        self.assertIn(new_pool, pools)

    @decorators.attr(type='smoke')
    def test_list_pools_two(self):
        """Test get pools with two pools"""
        new_pool1 = self._prepare_and_create_pool()
        new_pool2 = self._prepare_and_create_pool()
        pools = self.pools_client.list_pools()
        self.assertEqual(2, len(pools))
        self._prep_list_comparison(new_pool1, pools)
        self._prep_list_comparison(new_pool2, pools)
        self.assertIn(new_pool1, pools)
        self.assertIn(new_pool2, pools)

    @decorators.attr(type='smoke')
    def test_create_and_get_pool(self):
        """Test create pool"""
        new_pool = self._prepare_and_create_pool()
        pool = self.pools_client.get_pool(new_pool.get('id'))
        self._test_provisioning_status_if_exists(new_pool, pool)
        self.assertEqual(new_pool, pool)

    @decorators.attr(type='negative')
    def test_create_pool_missing_required_fields(self):
        """Test create pool with a missing required fields"""
        tenant_id = self.subnet.get('tenant_id')
        self.assertRaises(ex.BadRequest, self._create_pool,
                          tenant_id=tenant_id,
                          lb_algorithm='ROUND_ROBIN')

    def test_create_pool_missing_tenant_field(self):
        """Test create pool with a missing required tenant field"""
        tenant_id = self.subnet.get('tenant_id')
        new_pool = self._prepare_and_create_pool()
        pool = self.pools_client.get_pool(new_pool.get('id'))
        pool_tenant = pool['tenant_id']
        self.assertEqual(tenant_id, pool_tenant)

    @decorators.attr(type='negative')
    def test_create_pool_missing_protocol_field(self):
        """Test create pool with a missing required protocol field"""
        self.increment_protocol_port()
        listener = self.listeners_client.create_listener(
            loadbalancer_id=self.load_balancer.get('id'),
            protocol=self.listener_protocol, protocol_port=PROTOCOL_PORT)
        self.addCleanup(self._delete_listener, listener['id'])
        self._wait_for_load_balancer_status(self.load_balancer.get('id'))
        listener_id = listener.get('id')
        tenant_id = self.subnet.get('tenant_id')
        self.assertRaises(ex.BadRequest, self._create_pool,
                          tenant_id=tenant_id,
                          listener_id=listener_id,
                          lb_algorithm='ROUND_ROBIN')

    @decorators.attr(type='negative')
    def test_create_pool_missing_lb_algorithm_field(self):
        """Test create pool with a missing required lb algorithm field"""
        self.increment_protocol_port()
        listener = self.listeners_client.create_listener(
            loadbalancer_id=self.load_balancer.get('id'),
            protocol=self.listener_protocol, protocol_port=PROTOCOL_PORT)
        self.addCleanup(self._delete_listener, listener['id'])
        self._wait_for_load_balancer_status(self.load_balancer.get('id'))
        listener_id = listener.get('id')
        tenant_id = self.subnet.get('tenant_id')
        self.assertRaises(ex.BadRequest, self._create_pool,
                          tenant_id=tenant_id,
                          listener_id=listener_id,
                          protocol=self.pool_protocol)

    @decorators.attr(type='negative')
    def test_create_pool_missing_listener_id_field(self):
        """Test create pool with a missing required listener id field"""
        tenant_id = self.subnet.get('tenant_id')
        self.assertRaises(ex.BadRequest, self._create_pool,
                          tenant_id=tenant_id,
                          lb_algorithm='ROUND_ROBIN',
                          protocol=self.pool_protocol)

    def test_create_pool_missing_description_field(self):
        """Test create pool with missing description field"""
        self._wait_for_load_balancer_status(self.load_balancer.get('id'))
        new_pool = self._prepare_and_create_pool()
        pool_initial = self.pools_client.get_pool(new_pool.get('id'))
        desc = pool_initial.get('description')
        self.assertEqual(desc, "")

    def test_create_pool_missing_name_field(self):
        """Test create pool with a missing name field"""
        new_pool = self._prepare_and_create_pool()
        pool_initial = self.pools_client.get_pool(new_pool.get('id'))
        name = pool_initial.get('name')
        self.assertEqual(name, "")

    def test_create_pool_missing_admin_state_up_field(self):
        """Test create pool with a missing admin_state_up field"""
        new_pool = self._prepare_and_create_pool()
        pool_initial = self.pools_client.get_pool(new_pool.get('id'))
        state = pool_initial.get('admin_state_up')
        self.assertEqual(state, True)

    def test_create_pool_missing_session_pers_field(self):
        """Test create pool with a missing session_pers field"""
        new_pool = self._prepare_and_create_pool()
        pool_initial = self.pools_client.get_pool(new_pool.get('id'))
        sess = pool_initial.get('session_persistence')
        self.assertIsNone(sess)

    @decorators.attr(type='negative')
    def test_create_pool_invalid_protocol(self):
        """Test create pool with an invalid protocol"""
        self.assertRaises(ex.BadRequest, self._create_pool,
                          protocol='UDP',
                          lb_algorithm='ROUND_ROBIN',
                          listener_id=self.listener['id'])

    @decorators.attr(type='negative')
    def test_create_pool_invalid_session_persistence_field(self):
        """Test create pool with invalid session persistence field"""
        self.assertRaises(ex.BadRequest, self._create_pool,
                          protocol=self.pool_protocol,
                          session_persistence={'type': 'HTTP'},
                          lb_algorithm='ROUND_ROBIN',
                          listener_id=self.listener['id'])

    @decorators.attr(type='negative')
    def test_create_pool_invalid_algorithm(self):
        """Test create pool with an invalid algorithm"""
        self.assertRaises(ex.BadRequest, self._create_pool,
                          protocol=self.pool_protocol,
                          lb_algorithm='LEAST_CON',
                          listener_id=self.listener['id'])

    @decorators.attr(type='negative')
    def test_create_pool_invalid_admin_state_up(self):
        """Test create pool with an invalid admin state up field"""
        self.assertRaises(ex.BadRequest, self._create_pool,
                          protocol=self.pool_protocol,
                          admin_state_up="$!1%9823",
                          lb_algorithm='ROUND_ROBIN',
                          listener_id=self.listener['id'])

    @decorators.attr(type='negative')
    def test_create_pool_invalid_listener_field(self):
        """Test create pool with invalid listener field"""
        tenant_id = self.subnet.get('tenant_id')
        self.assertRaises(ex.BadRequest, self._create_pool,
                          tenant_id=tenant_id,
                          lb_algorithm='ROUND_ROBIN',
                          protocol=self.pool_protocol,
                          listener_id="$@5$%$7863")

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_create_pool_invalid_tenant_id_field(self):
        """Test create pool with invalid tenant_id field"""
        self.increment_protocol_port()
        listener = self.listeners_client.create_listener(
            loadbalancer_id=self.load_balancer.get('id'),
            protocol=self.listener_protocol, protocol_port=PROTOCOL_PORT)
        self.addCleanup(self._delete_listener, listener['id'])
        self._wait_for_load_balancer_status(self.load_balancer.get('id'))
        listener_id = listener.get('id')
        self.assertRaises(ex.BadRequest, self._create_pool,
                          tenant_id="*&7653^%&",
                          lb_algorithm='ROUND_ROBIN',
                          protocol=self.pool_protocol,
                          listener_id=listener_id)

    @decorators.attr(type='negative')
    def test_create_pool_incorrect_attribute(self):
        """Test create a pool with an extra, incorrect field"""
        self.assertRaises(ex.BadRequest, self._create_pool,
                          protocol=self.pool_protocol,
                          lb_algorithm='ROUND_ROBIN',
                          protocol_port=80,
                          listener_id=self.listener['id'])

    @decorators.attr(type='negative')
    def test_create_pool_empty_listener_field(self):
        """Test create pool with empty listener field"""
        tenant_id = self.subnet.get('tenant_id')
        self.assertRaises(ex.BadRequest, self._create_pool,
                          tenant_id=tenant_id,
                          lb_algorithm='ROUND_ROBIN',
                          protocol=self.pool_protocol,
                          listener_id="")

    def test_create_pool_empty_description_field(self):
        """Test create pool with empty description field"""
        new_pool = self._prepare_and_create_pool(description="")
        pool = self.pools_client.get_pool(new_pool.get('id'))
        pool_desc = pool.get('description')
        self.assertEqual(pool_desc, '')

    def test_create_pool_empty_name_field(self):
        """Test create pool with empty name field"""
        new_pool = self._prepare_and_create_pool(name="")
        pool = self.pools_client.get_pool(new_pool.get('id'))
        pool_name = pool.get('name')
        self.assertEqual(pool_name, '')

    @decorators.attr(type='negative')
    def test_create_pool_empty_protocol(self):
        """Test create pool with an empty protocol"""
        self.assertRaises(ex.BadRequest, self._create_pool,
                          protocol="",
                          lb_algorithm='ROUND_ROBIN',
                          listener_id=self.listener['id'])

    @decorators.skip_because(bug="1765796")
    @decorators.attr(type='negative')
    def test_create_pool_empty_session_persistence_field(self):
        """Test create pool with empty session persistence field"""
        self.assertRaises(ex.BadRequest, self._create_pool,
                          session_persistence="",
                          protocol=self.pool_protocol,
                          lb_algorithm='ROUND_ROBIN',
                          listener_id=self.listener['id'])

    @decorators.attr(type='negative')
    def test_create_pool_empty_algorithm(self):
        """Test create pool with an empty algorithm"""
        self.assertRaises(ex.BadRequest, self._create_pool,
                          protocol=self.pool_protocol,
                          lb_algorithm="",
                          listener_id=self.listener['id'])

    @decorators.attr(type='negative')
    def test_create_pool_empty_admin_state_up(self):
        """Test create pool with an invalid admin state up field"""
        self.assertRaises(ex.BadRequest, self._create_pool,
                          protocol=self.pool_protocol,
                          admin_state_up="",
                          lb_algorithm='ROUND_ROBIN')

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_create_pool_empty_tenant_field(self):
        """Test create pool with empty tenant field"""
        self.assertRaises(ex.BadRequest, self._create_pool,
                          protocol=self.pool_protocol,
                          tenant_id="",
                          lb_algorithm='ROUND_ROBIN',
                          listener_id=self.listener['id'])

    @decorators.skip_because(bug="1468457")
    @decorators.attr(type='negative')
    def test_create_pool_for_other_tenant_field(self):
        """Test create pool for other tenant field"""
        tenant = 'deffb4d7c0584e89a8ec99551565713c'
        self.assertRaises(ex.BadRequest, self._create_pool,
                          protocol=self.pool_protocol,
                          tenant_id=tenant,
                          lb_algorithm='ROUND_ROBIN',
                          listener_id=self.listener['id'])

    @decorators.attr(type='negative')
    def test_create_pool_invalid_name_field(self):
        """
        known bug with input more than 255 chars
        Test create pool with invalid name field
        """
        self.assertRaises(ex.BadRequest, self._create_pool,
                          protocol=self.pool_protocol,
                          lb_algorithm='ROUND_ROBIN',
                          listener_id=self.listener['id'],
                          name='n' * 256)

    @decorators.attr(type='negative')
    def test_create_pool_invalid_desc_field(self):
        """
        known bug with input more than 255 chars
        Test create pool with invalid desc field
        """
        self.assertRaises(ex.BadRequest, self._prepare_and_create_pool,
                          protocol=self.pool_protocol,
                          lb_algorithm='ROUND_ROBIN',
                          listener_id=self.listener['id'],
                          description='d' * 256)

    @decorators.attr(type='negative')
    def test_create_pool_with_session_persistence_unsupported_type(self):
        """Test create a pool with an incorrect type value
        for session persistence
        """
        self.assertRaises(ex.BadRequest, self._create_pool,
                          session_persistence={'type': 'UNSUPPORTED'},
                          protocol=self.pool_protocol,
                          lb_algorithm='ROUND_ROBIN',
                          listener_id=self.listener['id'])

    @decorators.attr(type='smoke')
    def test_create_pool_with_session_persistence_http_cookie(self):
        """Test create a pool with session_persistence type=HTTP_COOKIE"""
        new_pool = self._prepare_and_create_pool(
            session_persistence={'type': 'HTTP_COOKIE'})
        pool = self.pools_client.get_pool(new_pool.get('id'))
        self._test_provisioning_status_if_exists(new_pool, pool)
        self.assertEqual(new_pool, pool)

    def test_create_pool_with_session_persistence_app_cookie(self):
        """Test create a pool with session_persistence type=APP_COOKIE"""
        new_pool = self._prepare_and_create_pool(
            session_persistence={'type': 'APP_COOKIE',
                                 'cookie_name': 'sessionId'})
        pool = self.pools_client.get_pool(new_pool.get('id'))
        self._test_provisioning_status_if_exists(new_pool, pool)
        self.assertEqual(new_pool, pool)

    @decorators.attr(type='negative')
    def test_create_pool_with_session_persistence_redundant_cookie_name(self):
        """Test create a pool with session_persistence with cookie_name
        for type=HTTP_COOKIE
        """
        self.assertRaises(ex.BadRequest, self._create_pool,
                          session_persistence={'type': 'HTTP_COOKIE',
                                               'cookie_name': 'sessionId'},
                          protocol=self.pool_protocol,
                          lb_algorithm='ROUND_ROBIN',
                          listener_id=self.listener['id'])

    @decorators.attr(type='negative')
    def test_create_pool_with_session_persistence_without_cookie_name(self):
        """Test create a pool with session_persistence without
        cookie_name for type=APP_COOKIE
        """
        self.assertRaises(ex.BadRequest, self._create_pool,
                          session_persistence={'type': 'APP_COOKIE'},
                          protocol=self.pool_protocol,
                          lb_algorithm='ROUND_ROBIN',
                          listener_id=self.listener['id'])

    @decorators.attr(type='smoke')
    def test_update_pool(self):
        """Test update pool"""
        new_pool = self._prepare_and_create_pool()
        desc = 'testing update with new description'
        self._update_pool(new_pool.get('id'), description=desc, wait=True)
        pool = self.pools_client.get_pool(new_pool.get('id'))
        self.assertEqual(desc, pool.get('description'))

    def test_update_pool_missing_name(self):
        """Test update pool with missing name"""
        new_pool = self._prepare_and_create_pool()
        pool_initial = self.pools_client.get_pool(new_pool.get('id'))
        name = pool_initial.get('name')
        pool = self.pools_client.update_pool(new_pool.get('id'))
        self._wait_for_load_balancer_status(self.load_balancer.get('id'))
        self.assertEqual(name, pool.get('name'))

    def test_update_pool_missing_description(self):
        """Test update pool with missing description"""
        new_pool = self._prepare_and_create_pool()
        pool_initial = self.pools_client.get_pool(new_pool.get('id'))
        desc = pool_initial.get('description')
        pool = self.pools_client.update_pool(new_pool.get('id'))
        self._wait_for_load_balancer_status(self.load_balancer.get('id'))
        self.assertEqual(desc, pool.get('description'))

    def test_update_pool_missing_admin_state_up(self):
        """Test update pool with missing admin state up field"""
        new_pool = self._prepare_and_create_pool()
        pool_initial = self.pools_client.get_pool(new_pool.get('id'))
        admin = pool_initial.get('admin_state_up')
        pool = self.pools_client.update_pool(new_pool.get('id'))
        self._wait_for_load_balancer_status(self.load_balancer.get('id'))
        self.assertEqual(admin, pool.get('admin_state_up'))

    def test_update_pool_missing_session_persistence(self):
        """Test update pool with missing session persistence"""
        new_pool = self._prepare_and_create_pool()
        pool_initial = self.pools_client.get_pool(new_pool.get('id'))
        sess_pers = pool_initial.get('session_persistence')
        pool = self._update_pool(new_pool.get('id'))
        self.assertAlmostEqual(sess_pers, pool.get('session_persistence'))

    @decorators.attr(type='negative')
    def test_update_pool_invalid_name(self):
        """Test update pool with invalid name"""
        new_pool = self._prepare_and_create_pool()
        self.assertRaises(ex.BadRequest, self.pools_client.update_pool,
                          new_pool.get('id'), name='n' * 256)

    @decorators.attr(type='negative')
    def test_update_pool_invalid_desc(self):
        """Test update pool with invalid desc"""
        new_pool = self._prepare_and_create_pool()
        self.assertRaises(ex.BadRequest, self.pools_client.update_pool,
                          new_pool.get('id'),
                          description='d' * 256)

    @decorators.attr(type='negative')
    def test_update_pool_invalid_admin_state_up(self):
        """Test update pool with an invalid admin_state_up"""
        new_pool = self._prepare_and_create_pool()
        self.assertRaises(ex.BadRequest, self.pools_client.update_pool,
                          new_pool.get('id'), admin_state_up='hello')

    @decorators.attr(type='negative')
    def test_update_pool_invalid_session_persistence(self):
        """Test update pool with an invalid session pers. field"""
        new_pool = self._prepare_and_create_pool()
        self.assertRaises(ex.BadRequest, self.pools_client.update_pool,
                          new_pool.get('id'),
                          session_persistence={'type': 'Hello'})

    def test_update_pool_empty_name(self):
        """Test update pool with empty name"""
        new_pool = self._prepare_and_create_pool()
        pool = self.pools_client.update_pool(new_pool.get('id'), name="")
        self._wait_for_load_balancer_status(self.load_balancer.get('id'))
        self.assertEqual(pool.get('name'), "")

    def test_update_pool_empty_description(self):
        """Test update pool with empty description"""
        new_pool = self._prepare_and_create_pool()
        pool = self.pools_client.update_pool(new_pool.get('id'),
                                             description="")
        self._wait_for_load_balancer_status(self.load_balancer.get('id'))
        self.assertEqual(pool.get('description'), "")

    @decorators.attr(type='negative')
    def test_update_pool_empty_admin_state_up(self):
        """Test update pool with empty admin state up"""
        new_pool = self._prepare_and_create_pool()
        self.assertRaises(ex.BadRequest, self.pools_client.update_pool,
                          new_pool.get('id'), admin_state_up="")

    @decorators.skip_because(bug="1765796")
    @decorators.attr(type='negative')
    def test_update_pool_empty_session_persistence(self):
        """Test update pool with empty session persistence field"""
        new_pool = self._prepare_and_create_pool()
        self.assertRaises(ex.BadRequest, self.pools_client.update_pool,
                          new_pool.get('id'),
                          session_persistence="")

    @decorators.attr(type='negative')
    def test_update_pool_invalid_attribute(self):
        """Test update pool with an invalid attribute"""
        new_pool = self._prepare_and_create_pool()
        self.assertRaises(ex.BadRequest, self._update_pool,
                          new_pool.get('id'), lb_algorithm='ROUNDED')

    @decorators.attr(type='negative')
    def test_update_pool_incorrect_attribute(self):
        """Test update a pool with an extra, incorrect field"""
        new_pool = self._prepare_and_create_pool()
        self.assertRaises(ex.BadRequest, self._update_pool,
                          new_pool.get('id'), protocol='HTTPS')

    @decorators.attr(type='smoke')
    def test_delete_pool(self):
        """Test delete pool"""
        new_pool = self._prepare_and_create_pool(cleanup=False)
        pool = self.pools_client.get_pool(new_pool.get('id'))
        self._test_provisioning_status_if_exists(new_pool, pool)
        self.assertEqual(new_pool, pool)
        self._delete_pool(new_pool.get('id'))
        self.assertRaises(ex.NotFound, self.pools_client.get_pool,
                          new_pool.get('id'))

    def test_delete_invalid_pool(self):
        """Test delete pool that doesn't exist"""
        new_pool = self._prepare_and_create_pool(cleanup=False)
        pool = self.pools_client.get_pool(new_pool.get('id'))
        self._test_provisioning_status_if_exists(new_pool, pool)
        self.assertEqual(new_pool, pool)
        self._delete_pool(new_pool.get('id'))
        self.assertRaises(ex.NotFound, self._delete_pool,
                          new_pool.get('id'))



# Copyright 2013 OpenStack Foundation.  All rights reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Neutron Lbaas base exception handling.
"""

from neutron_lib import exceptions

from neutron_lbaas._i18n import _


class ModelMapException(exceptions.NeutronException):
    message = _("Unable to map model class %(target_name)s")


class LbaasException(exceptions.NeutronException):
    pass


class TLSException(LbaasException):
    pass


class NeedsPassphrase(TLSException):
    message = _("Passphrase needed to decrypt key but client "
               "did not provide one.")


class UnreadableCert(TLSException):
    message = _("Could not read X509 from PEM")


class MisMatchedKey(TLSException):
    message = _("Key and x509 certificate do not match")


class CertificateStorageException(TLSException):
    message = _('Could not store certificate: %(msg)s')


class LoadbalancerReschedulingFailed(exceptions.Conflict):
    message = _("Failed rescheduling loadbalancer %(loadbalancer_id)s: "
                "no eligible lbaas agent found.")


class BadRequestException(exceptions.BadRequest):
    message = "%(fault_string)s"


class ConflictException(exceptions.Conflict):
    message = "%(fault_string)s"


class NotAuthorizedException(exceptions.NotAuthorized):
    message = "%(fault_string)s"


class NotFoundException(exceptions.NotFound):
    message = "%(fault_string)s"


class ServiceUnavailableException(exceptions.ServiceUnavailable):
    message = "%(fault_string)s"


class UnknownException(exceptions.NeutronException):
    message = "%(fault_string)s"



# Copyright 2015 OpenStack Foundation
# Copyright 2015 Blue Box, an IBM Company
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""independent pools

Revision ID: 6aee0434f911
Revises: 3426acbc12de
Create Date: 2015-08-28 03:15:42.533386

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '6aee0434f911'
down_revision = '3426acbc12de'


def upgrade():
    conn = op.get_bind()
    # Minimal examples of the tables we need to manipulate
    listeners = sa.sql.table(
        'lbaas_listeners',
        sa.sql.column('loadbalancer_id', sa.String),
        sa.sql.column('default_pool_id', sa.String))
    pools = sa.sql.table(
        'lbaas_pools',
        sa.sql.column('loadbalancer_id', sa.String),
        sa.sql.column('id', sa.String))

    # This foreign key does not need to be unique anymore. To remove the
    # uniqueness but keep the foreign key we have to do some juggling.
    #
    # Also, because different database engines handle unique constraints
    # in incompatible ways, we can't simply call op.drop_constraint and
    # expect it to work for all DB engines. This is yet another unfortunate
    # case where sqlalchemy isn't able to abstract everything away.
    if op.get_context().dialect.name == 'postgresql':
        # PostgreSQL path:
        op.drop_constraint('lbaas_listeners_default_pool_id_key',
                           'lbaas_listeners', 'unique')
    else:
        # MySQL path:
        op.drop_constraint('lbaas_listeners_ibfk_2', 'lbaas_listeners',
                           type_='foreignkey')
        op.drop_constraint('default_pool_id', 'lbaas_listeners',
                           type_='unique')
        op.create_foreign_key('lbaas_listeners_ibfk_2', 'lbaas_listeners',
                              'lbaas_pools', ['default_pool_id'], ['id'])

    op.add_column(
        u'lbaas_pools',
        sa.Column('loadbalancer_id', sa.String(36),
            sa.ForeignKey('lbaas_loadbalancers.id'), nullable=True)
    )

    # Populate this new column appropriately
    select_obj = sa.select([listeners.c.loadbalancer_id,
                           listeners.c.default_pool_id]).where(
                           listeners.c.default_pool_id is not None)
    result = conn.execute(select_obj)
    for row in result:
        stmt = pools.update().values(loadbalancer_id=row[0]).where(
            pools.c.id == row[1])
        op.execute(stmt)

# For existing installations, the above ETL should populate the above column
# using the following procedure:
#
# Get the output from this:
#
# SELECT default_pool_id, loadbalancer_id l_id FROM lbaas_listeners WHERE
# default_pool_id IS NOT NULL;
#
# Then for every row returned run:
#
# UPDATE lbaas_pools SET loadbalancer_id = l_id WHERE id = default_pool_id;



# Copyright 2015, 2016 Rackspace Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from neutron_lib import constants as n_constants
from oslo_log import log as logging
from tempest.api.network import base
from tempest import config
from tempest.lib.common.utils import test_utils
from tempest.lib import exceptions

from neutron_lbaas._i18n import _
from neutron_lbaas.tests.tempest.v2.clients import health_monitors_client
from neutron_lbaas.tests.tempest.v2.clients import listeners_client
from neutron_lbaas.tests.tempest.v2.clients import load_balancers_client
from neutron_lbaas.tests.tempest.v2.clients import members_client
from neutron_lbaas.tests.tempest.v2.clients import pools_client

CONF = config.CONF

LOG = logging.getLogger(__name__)


def _setup_client_args(auth_provider):
    """Set up ServiceClient arguments using config settings. """
    service = CONF.network.catalog_type or 'network'
    region = CONF.network.region or 'regionOne'
    endpoint_type = CONF.network.endpoint_type
    build_interval = CONF.network.build_interval
    build_timeout = CONF.network.build_timeout

    # The disable_ssl appears in identity
    disable_ssl_certificate_validation = (
        CONF.identity.disable_ssl_certificate_validation)
    ca_certs = None

    # Trace in debug section
    trace_requests = CONF.debug.trace_requests

    return [auth_provider, service, region, endpoint_type,
            build_interval, build_timeout,
            disable_ssl_certificate_validation, ca_certs,
            trace_requests]


class BaseTestCase(base.BaseNetworkTest):

    # This class picks non-admin credentials and run the tempest tests

    _lbs_to_delete = []

    @classmethod
    def resource_setup(cls):
        super(BaseTestCase, cls).resource_setup()

        cls.listener_protocol = CONF.lbaas.default_listener_protocol
        cls.pool_protocol = CONF.lbaas.default_pool_protocol
        cls.hm_protocol = CONF.lbaas.default_health_monitor_protocol

        mgr = cls.get_client_manager()
        auth_provider = mgr.auth_provider
        client_args = _setup_client_args(auth_provider)

        cls.load_balancers_client = (
            load_balancers_client.LoadBalancersClientJSON(*client_args))
        cls.listeners_client = (
            listeners_client.ListenersClientJSON(*client_args))
        cls.pools_client = pools_client.PoolsClientJSON(*client_args)
        cls.members_client = members_client.MembersClientJSON(*client_args)
        cls.health_monitors_client = (
            health_monitors_client.HealthMonitorsClientJSON(*client_args))

    @classmethod
    def resource_cleanup(cls):

        for lb_id in cls._lbs_to_delete:
            try:
                lb = cls.load_balancers_client.get_load_balancer_status_tree(
                    lb_id).get('loadbalancer')
            except exceptions.NotFound:
                continue
            for listener in lb.get('listeners'):
                for pool in listener.get('pools'):
                    # delete pool's health-monitor
                    hm = pool.get('healthmonitor')
                    if hm:
                        test_utils.call_and_ignore_notfound_exc(
                            cls.health_monitors_client.delete_health_monitor,
                            pool.get('healthmonitor').get('id'))
                        cls._wait_for_load_balancer_status(lb_id)
                    test_utils.call_and_ignore_notfound_exc(
                        cls.pools_client.delete_pool,
                        pool.get('id'))
                    cls._wait_for_load_balancer_status(lb_id)
                    # delete pool's members
                    members = pool.get('members', [])
                    for member in members:
                        test_utils.call_and_ignore_notfound_exc(
                            cls.members_client.delete_member,
                            pool.get('id'), member.get('id'))
                        cls._wait_for_load_balancer_status(lb_id)
                    # delete pool
                    test_utils.call_and_ignore_notfound_exc(
                        cls.pools_client.delete_pool, pool.get('id'))
                    cls._wait_for_load_balancer_status(lb_id)
                # delete listener
                test_utils.call_and_ignore_notfound_exc(
                    cls.listeners_client.delete_listener,
                    listener.get('id'))
                cls._wait_for_load_balancer_status(lb_id)
            # delete load-balancer
            test_utils.call_and_ignore_notfound_exc(
                cls._delete_load_balancer, lb_id)

        # Wait for all LBs to be fully deleted
        for lb_id in cls._lbs_to_delete:
            cls._wait_for_load_balancer_status(
                lb_id, provisioning_status='DELETED', operating_status=None,
                delete=True)

        # Wait for straggling port deletes to finish (neutron will sometimes
        # claim they are deleted before they are totally finished deleting)
        remaining_ports = cls.ports_client.list_ports(
            network_id=cls.network['id']).get('ports')
        tries = 1
        while remaining_ports and tries <= 10:
            LOG.warning("Ports still exist on network %s (try #%s): %s",
                        cls.network['id'], tries, remaining_ports)
            time.sleep(10)
            remaining_ports = cls.ports_client.list_ports(
                network_id=cls.network['id']).get('ports')
            tries += 1

        if remaining_ports:
            LOG.error("Failed to remove all remaining ports on network %s. "
                      "Subnet deletion will probably fail.", cls.network['id'])

        super(BaseTestCase, cls).resource_cleanup()

    @classmethod
    def setUpClass(cls):
        cls.LOG = logging.getLogger(cls._get_full_case_name())
        super(BaseTestCase, cls).setUpClass()

    def setUp(cls):
        cls.LOG.info('Starting: {0}'.format(cls._testMethodName))
        super(BaseTestCase, cls).setUp()

    def tearDown(cls):
        super(BaseTestCase, cls).tearDown()
        cls.LOG.info('Finished: {0}\n'.format(cls._testMethodName))

    @classmethod
    def _create_load_balancer(cls, wait=True, **lb_kwargs):
        lb = cls.load_balancers_client.create_load_balancer(**lb_kwargs)
        if wait:
            cls._wait_for_load_balancer_status(lb.get('id'))

        cls._lbs_to_delete.append(lb.get('id'))
        return lb

    @classmethod
    def _create_active_load_balancer(cls, **kwargs):
        lb = cls._create_load_balancer(**kwargs)
        lb = cls._wait_for_load_balancer_status(lb.get('id'))
        return lb

    @classmethod
    def _delete_load_balancer(cls, load_balancer_id, wait=True):
        cls.load_balancers_client.delete_load_balancer(load_balancer_id)
        if wait:
            cls._wait_for_load_balancer_status(
                load_balancer_id, delete=True)

    @classmethod
    def _update_load_balancer(cls, load_balancer_id, wait=True, **lb_kwargs):
        lb = cls.load_balancers_client.update_load_balancer(
            load_balancer_id, **lb_kwargs)
        if wait:
            cls._wait_for_load_balancer_status(
                load_balancer_id)
        return lb

    @classmethod
    def _wait_for_load_balancer_status(cls, load_balancer_id,
                                       provisioning_status='ACTIVE',
                                       operating_status='ONLINE',
                                       delete=False):
        interval_time = 1
        timeout = 600
        end_time = time.time() + timeout
        lb = {}
        # When running with no-op drivers there is no actual health to
        # observe, so disable operating_status checks when running no-op.
        if CONF.lbaas.test_with_noop:
            operating_status = None
        while time.time() < end_time:
            try:
                lb = cls.load_balancers_client.get_load_balancer(
                    load_balancer_id)
                if not lb:
                        # loadbalancer not found
                    if delete:
                        break
                    else:
                        raise Exception(
                            _("loadbalancer {lb_id} not"
                              " found").format(
                                  lb_id=load_balancer_id))
                if lb.get('provisioning_status') == provisioning_status:
                    if operating_status is None:
                        break
                    elif lb.get('operating_status') == operating_status:
                        break
                time.sleep(interval_time)
            except exceptions.NotFound:
                # if wait is for delete operation do break
                if delete:
                    break
                else:
                    # raise original exception
                    raise
        else:
            if delete:
                raise exceptions.TimeoutException(
                    _("Waited for load balancer {lb_id} to be deleted for "
                      "{timeout} seconds but can still observe that it "
                      "exists.").format(
                          lb_id=load_balancer_id,
                          timeout=timeout))
            else:
                raise exceptions.TimeoutException(
                    _("Wait for load balancer ran for {timeout} seconds and "
                      "did not observe {lb_id} reach {provisioning_status} "
                      "provisioning status and {operating_status} "
                      "operating status.").format(
                          timeout=timeout,
                          lb_id=load_balancer_id,
                          provisioning_status=provisioning_status,
                          operating_status=operating_status))
        return lb

    @classmethod
    def _wait_for_neutron_port_delete(cls, port_id):
        """
        Neutron port deletion is asynchronous. This method waits, up to a
        timeout for the port deletion to complete.
        """
        interval_time = 1
        timeout = 600
        end_time = time.time() + timeout
        while time.time() < end_time:
            try:
                cls.ports_client.show_port(port_id)
                time.sleep(interval_time)
            except exceptions.NotFound:
                break
        else:
            raise exceptions.TimeoutException(
                    _("Waited for port {port_id} to be deleted for "
                      "{timeout} seconds but can still observe that it "
                      "exists.").format(port_id=port_id, timeout=timeout))

    @classmethod
    def _create_listener(cls, wait=True, **listener_kwargs):
        listener = cls.listeners_client.create_listener(**listener_kwargs)
        if wait:
            cls._wait_for_load_balancer_status(cls.load_balancer.get('id'))
        return listener

    @classmethod
    def _delete_listener(cls, listener_id, wait=True):
        cls.listeners_client.delete_listener(listener_id)
        if wait:
            cls._wait_for_load_balancer_status(cls.load_balancer.get('id'))

    @classmethod
    def _update_listener(cls, listener_id, wait=True, **listener_kwargs):
        listener = cls.listeners_client.update_listener(
            listener_id, **listener_kwargs)
        if wait:
            cls._wait_for_load_balancer_status(
                cls.load_balancer.get('id'))
        return listener

    @classmethod
    def _create_pool(cls, wait=True, **pool_kwargs):
        pool = cls.pools_client.create_pool(**pool_kwargs)
        if wait:
            cls._wait_for_load_balancer_status(cls.load_balancer.get('id'))
        return pool

    @classmethod
    def _delete_pool(cls, pool_id, wait=True):
        cls.pools_client.delete_pool(pool_id)
        if wait:
            cls._wait_for_load_balancer_status(cls.load_balancer.get('id'))

    @classmethod
    def _update_pool(cls, pool_id, wait=True, **pool_kwargs):
        pool = cls.pools_client.update_pool(pool_id, **pool_kwargs)
        if wait:
            cls._wait_for_load_balancer_status(
                cls.load_balancer.get('id'))
        return pool

    def _create_health_monitor(self, wait=True, cleanup=True,
                               **health_monitor_kwargs):
        hm = self.health_monitors_client.create_health_monitor(
            **health_monitor_kwargs)
        if cleanup:
            self.addCleanup(self._delete_health_monitor, hm.get('id'))
        if wait:
            self._wait_for_load_balancer_status(self.load_balancer.get('id'))
        return hm

    def _delete_health_monitor(self, health_monitor_id, wait=True):
        self.health_monitors_client.delete_health_monitor(health_monitor_id)
        if wait:
            self._wait_for_load_balancer_status(self.load_balancer.get('id'))

    def _update_health_monitor(self, health_monitor_id, wait=True,
                               **health_monitor_kwargs):
        health_monitor = self.health_monitors_client.update_health_monitor(
            health_monitor_id, **health_monitor_kwargs)
        if wait:
            self._wait_for_load_balancer_status(
                self.load_balancer.get('id'))
        return health_monitor

    @classmethod
    def _create_member(cls, pool_id, wait=True, **member_kwargs):
        member = cls.members_client.create_member(pool_id, **member_kwargs)
        if wait:
            cls._wait_for_load_balancer_status(cls.load_balancer.get('id'))
        return member

    @classmethod
    def _delete_member(cls, pool_id, member_id, wait=True):
        cls.members_client.delete_member(pool_id, member_id)
        if wait:
            cls._wait_for_load_balancer_status(cls.load_balancer.get('id'))

    @classmethod
    def _update_member(cls, pool_id, member_id, wait=True,
                       **member_kwargs):
        member = cls.members_client.update_member(
            pool_id, member_id, **member_kwargs)
        if wait:
            cls._wait_for_load_balancer_status(
                cls.load_balancer.get('id'))
        return member

    @classmethod
    def _check_status_tree(cls, load_balancer_id, listener_ids=None,
                           pool_ids=None, health_monitor_id=None,
                           member_ids=None):
        statuses = cls.load_balancers_client.get_load_balancer_status_tree(
            load_balancer_id=load_balancer_id)
        load_balancer = statuses['loadbalancer']

        # When running with no-op drivers there is no actual health to
        # observe, so disable operating_status checks when running no-op.
        if not CONF.lbaas.test_with_noop:
            assert 'ONLINE' == load_balancer['operating_status']

        assert 'ACTIVE' == load_balancer['provisioning_status']

        if listener_ids:
            cls._check_status_tree_thing(listener_ids,
                                         load_balancer['listeners'])
        if pool_ids:
            cls._check_status_tree_thing(pool_ids,
                                         load_balancer['listeners']['pools'])
        if member_ids:
            cls._check_status_tree_thing(
                member_ids,
                load_balancer['listeners']['pools']['members'])
        if health_monitor_id:
            health_monitor = (
                load_balancer['listeners']['pools']['health_monitor'])
            assert health_monitor_id == health_monitor['id']
            assert 'ACTIVE' == health_monitor['provisioning_status']

    @classmethod
    def _check_status_tree_thing(cls, actual_thing_ids, status_tree_things):
        found_things = 0
        assert len(actual_thing_ids) == len(status_tree_things)
        for actual_thing_id in actual_thing_ids:
            for status_tree_thing in status_tree_things:
                if status_tree_thing['id'] == actual_thing_id:
                    # When running with no-op drivers there is no actual
                    # health to observe, so disable operating_status checks
                    # when running no-op.
                    if not CONF.lbaas.test_with_noop:
                        assert 'ONLINE' == (
                            status_tree_thing['operating_status'])
                    assert 'ACTIVE' == (
                        status_tree_thing['provisioning_status'])
                    found_things += 1
        assert len(actual_thing_ids) == found_things

    @classmethod
    def _get_full_case_name(cls):
        name = '{module}:{case_name}'.format(
            module=cls.__module__,
            case_name=cls.__name__
        )
        return name

    def _test_provisioning_status_if_exists(self, created_obj, get_obj):
        prov_status_old = created_obj.pop('provisioning_status', None)
        prov_status_new = get_obj.pop('provisioning_status', None)

        if prov_status_old:
            self.assertIn('PENDING_', prov_status_old)
        if prov_status_new:
            self.assertEqual(n_constants.ACTIVE, prov_status_new)
        if not created_obj.get('updated_at') and get_obj.get('updated_at'):
            get_obj['updated_at'] = None


class BaseAdminTestCase(BaseTestCase):

    # This class picks admin credentials and run the tempest tests

    @classmethod
    def resource_setup(cls):

        super(BaseAdminTestCase, cls).resource_setup()

        mgr = cls.get_client_manager(credential_type='admin')
        auth_provider_admin = mgr.auth_provider
        client_args = _setup_client_args(auth_provider_admin)

        cls.load_balancers_client = (
            load_balancers_client.LoadBalancersClientJSON(*client_args))
        cls.listeners_client = (
            listeners_client.ListenersClientJSON(*client_args))
        cls.pools_client = (
            pools_client.PoolsClientJSON(*client_args))
        cls.members_client = (
            members_client.MembersClientJSON(*client_args))
        cls.health_monitors_client = (
            health_monitors_client.HealthMonitorsClientJSON(*client_args))

    @classmethod
    def resource_cleanup(cls):
        super(BaseAdminTestCase, cls).resource_cleanup()



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""lbaas version 2 api

Revision ID: lbaasv2
Revises: start_neutron_lbaas
Create Date: 2014-06-18 10:50:15.606420

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'lbaasv2'
down_revision = 'start_neutron_lbaas'


listener_protocols = sa.Enum("HTTP", "HTTPS", "TCP",
                             name="listener_protocolsv2")
pool_protocols = sa.Enum("HTTP", "HTTPS", "TCP",
                         name="pool_protocolsv2")
sesssionpersistences_type = sa.Enum("SOURCE_IP", "HTTP_COOKIE", "APP_COOKIE",
                                    name="sesssionpersistences_typev2")
lb_algorithms = sa.Enum("ROUND_ROBIN", "LEAST_CONNECTIONS", "SOURCE_IP",
                        name="lb_algorithmsv2")
healthmonitors_type = sa.Enum("PING", "TCP", "HTTP", "HTTPS",
                              name="healthmonitors_typev2")


def upgrade():
    op.create_table(
        u'lbaas_healthmonitors',
        sa.Column(u'tenant_id', sa.String(255), nullable=True),
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'type', healthmonitors_type, nullable=False),
        sa.Column(u'delay', sa.Integer(), nullable=False),
        sa.Column(u'timeout', sa.Integer(), nullable=False),
        sa.Column(u'max_retries', sa.Integer(), nullable=False),
        sa.Column(u'http_method', sa.String(16), nullable=True),
        sa.Column(u'url_path', sa.String(255), nullable=True),
        sa.Column(u'expected_codes', sa.String(64), nullable=True),
        sa.Column(u'status', sa.String(16), nullable=False),
        sa.Column(u'admin_state_up', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint(u'id')
    )

    op.create_table(
        u'lbaas_pools',
        sa.Column(u'tenant_id', sa.String(255), nullable=True),
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'name', sa.String(255), nullable=True),
        sa.Column(u'description', sa.String(255), nullable=True),
        sa.Column(u'protocol', pool_protocols, nullable=False),
        sa.Column(u'lb_algorithm', lb_algorithms, nullable=False),
        sa.Column(u'healthmonitor_id', sa.String(36), nullable=True),
        sa.Column(u'status', sa.String(16), nullable=False),
        sa.Column(u'admin_state_up', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint(u'id'),
        sa.UniqueConstraint(u'healthmonitor_id'),
        sa.ForeignKeyConstraint([u'healthmonitor_id'],
                                [u'lbaas_healthmonitors.id'])
    )

    op.create_table(
        u'lbaas_sessionpersistences',
        sa.Column(u'pool_id', sa.String(36), nullable=False),
        sa.Column(u'type', sesssionpersistences_type, nullable=False),
        sa.Column(u'cookie_name', sa.String(1024), nullable=True),
        sa.ForeignKeyConstraint([u'pool_id'], [u'lbaas_pools.id']),
        sa.PrimaryKeyConstraint(u'pool_id')
    )

    op.create_table(
        u'lbaas_members',
        sa.Column(u'tenant_id', sa.String(255), nullable=True),
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'pool_id', sa.String(36), nullable=False),
        sa.Column(u'subnet_id', sa.String(36), nullable=True),
        sa.Column(u'address', sa.String(64), nullable=False),
        sa.Column(u'protocol_port', sa.Integer(), nullable=False),
        sa.Column(u'weight', sa.Integer(), nullable=True),
        sa.Column(u'status', sa.String(16), nullable=False),
        sa.Column(u'admin_state_up', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint(u'id'),
        sa.ForeignKeyConstraint([u'pool_id'], [u'lbaas_pools.id']),
        sa.UniqueConstraint(u'pool_id', u'address', u'protocol_port',
                            name=u'uniq_pool_address_port_v2')
    )

    op.create_table(
        u'lbaas_loadbalancers',
        sa.Column(u'tenant_id', sa.String(255), nullable=True),
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'name', sa.String(255), nullable=True),
        sa.Column(u'description', sa.String(255), nullable=True),
        sa.Column(u'vip_port_id', sa.String(36), nullable=True),
        sa.Column(u'vip_subnet_id', sa.String(36), nullable=False),
        sa.Column(u'vip_address', sa.String(36), nullable=True),
        sa.Column(u'status', sa.String(16), nullable=False),
        sa.Column(u'admin_state_up', sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint([u'vip_port_id'], [u'ports.id'],
                                name=u'fk_lbaas_loadbalancers_ports_id'),
        sa.PrimaryKeyConstraint(u'id')
    )

    op.create_table(
        u'lbaas_listeners',
        sa.Column(u'tenant_id', sa.String(255), nullable=True),
        sa.Column(u'id', sa.String(36), nullable=False),
        sa.Column(u'name', sa.String(255), nullable=True),
        sa.Column(u'description', sa.String(255), nullable=True),
        sa.Column(u'protocol', listener_protocols, nullable=False),
        sa.Column(u'protocol_port', sa.Integer(), nullable=False),
        sa.Column(u'connection_limit', sa.Integer(), nullable=True),
        sa.Column(u'loadbalancer_id', sa.String(36), nullable=True),
        sa.Column(u'default_pool_id', sa.String(36), nullable=True),
        sa.Column(u'status', sa.String(16), nullable=False),
        sa.Column(u'admin_state_up', sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint([u'loadbalancer_id'],
                                [u'lbaas_loadbalancers.id']),
        sa.ForeignKeyConstraint([u'default_pool_id'],
                                [u'lbaas_pools.id']),
        sa.UniqueConstraint(u'default_pool_id'),
        sa.UniqueConstraint(u'loadbalancer_id', u'protocol_port',
                            name=u'uniq_loadbalancer_listener_port'),
        sa.PrimaryKeyConstraint(u'id')
    )

    op.create_table(
        u'lbaas_loadbalancer_statistics',
        sa.Column(u'loadbalancer_id', sa.String(36), nullable=False),
        sa.Column(u'bytes_in', sa.BigInteger(), nullable=False),
        sa.Column(u'bytes_out', sa.BigInteger(), nullable=False),
        sa.Column(u'active_connections', sa.BigInteger(), nullable=False),
        sa.Column(u'total_connections', sa.BigInteger(), nullable=False),
        sa.PrimaryKeyConstraint(u'loadbalancer_id'),
        sa.ForeignKeyConstraint([u'loadbalancer_id'],
                                [u'lbaas_loadbalancers.id'])
    )



# Copyright 2015, Radware LTD. All rights reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import helpers as log_helpers

from neutron_lbaas._i18n import _
from neutron_lbaas.drivers import driver_base


VERSION = "P1.0.0"

driver_opts = [
    cfg.StrOpt('vdirect_address',
               help=_('IP address of vDirect server.')),
    cfg.StrOpt('ha_secondary_address',
               help=_('IP address of secondary vDirect server.')),
    cfg.StrOpt('vdirect_user',
               default='vDirect',
               help=_('vDirect user name.')),
    cfg.StrOpt('vdirect_password',
               default='radware',
               secret=True,
               help=_('vDirect user password.')),
    cfg.IntOpt('port',
               default=2189,
               help=_('vDirect port. Default:2189')),
    cfg.BoolOpt('ssl',
               default=True,
               help=_('Use SSL. Default: True')),
    cfg.BoolOpt('ssl_verify_context',
                default=True,
                help=_('Enables or disables the SSL context verification '
                       'for legacy python that verifies HTTPS '
                       'certificates by default. Default: True.')),
    cfg.IntOpt('timeout',
               default=5000,
               help=_('vDirect connection timeout. Default:5000')),
    cfg.StrOpt('base_uri',
               default='',
               help=_('Base vDirect URI. Default:\'\'')),
    cfg.StrOpt('service_adc_type',
               default="VA",
               help=_('Service ADC type. Default: VA.')),
    cfg.StrOpt('service_adc_version',
               default="",
               help=_('Service ADC version.')),
    cfg.BoolOpt('service_ha_pair',
                default=False,
                help=_('Enables or disables the Service HA pair. '
                       'Default: False.')),
    cfg.BoolOpt('configure_allowed_address_pairs',
                default=False,
                help=_('Enables or disables allowed address pairs '
                       'configuration for VIP addresses. '
                       'Default: False.')),
    cfg.IntOpt('service_throughput',
               default=1000,
               help=_('Service throughput. Default: 1000.')),
    cfg.IntOpt('service_ssl_throughput',
               default=100,
               help=_('Service SSL throughput. Default: 100.')),
    cfg.IntOpt('service_compression_throughput',
               default=100,
               help=_('Service compression throughput. Default: 100.')),
    cfg.IntOpt('service_cache',
               default=20,
               help=_('Size of service cache. Default: 20.')),
    cfg.ListOpt('service_resource_pool_ids',
                default=[],
                help=_('Resource pool IDs.')),
    cfg.IntOpt('service_isl_vlan',
               default=-1,
               help=_('A required VLAN for the interswitch link to use.')),
    cfg.BoolOpt('service_session_mirroring_enabled',
                default=False,
                help=_('Enable or disable Alteon interswitch link for '
                       'stateful session failover. Default: False.')),
    cfg.StrOpt('workflow_template_name',
               default='os_lb_v2',
               help=_('Name of the workflow template. Default: os_lb_v2.')),
    cfg.ListOpt('child_workflow_template_names',
               default=['manage_l3'],
               help=_('Name of child workflow templates used.'
                      'Default: manage_l3')),
    cfg.DictOpt('workflow_params',
                default={"twoleg_enabled": "_REPLACE_",
                         "ha_network_name": "HA-Network",
                         "ha_ip_pool_name": "default",
                         "allocate_ha_vrrp": True,
                         "allocate_ha_ips": True,
                         "data_port": 1,
                         "data_ip_address": "192.168.200.99",
                         "data_ip_mask": "255.255.255.0",
                         "gateway": "192.168.200.1",
                         "ha_port": 2},
                help=_('Parameter for l2_l3 workflow constructor.')),
    cfg.StrOpt('workflow_action_name',
               default='apply',
               help=_('Name of the workflow action. '
                      'Default: apply.')),
    cfg.StrOpt('stats_action_name',
               default='stats',
               help=_('Name of the workflow action for statistics. '
                      'Default: stats.'))
]

driver_debug_opts = [
    cfg.BoolOpt('provision_service',
                default=True,
                help=_('Provision ADC service?')),
    cfg.BoolOpt('configure_l3',
                default=True,
                help=_('Configule ADC with L3 parameters?')),
    cfg.BoolOpt('configure_l4',
                default=True,
                help=_('Configule ADC with L4 parameters?'))
]

cfg.CONF.register_opts(driver_opts, "radwarev2")
cfg.CONF.register_opts(driver_debug_opts, "radwarev2_debug")


class RadwareLBaaSBaseV2Driver(driver_base.LoadBalancerBaseDriver):

    def __init__(self, plugin):
        super(RadwareLBaaSBaseV2Driver, self).__init__(plugin)

        self.load_balancer = LoadBalancerManager(self)
        self.listener = ListenerManager(self)
        self.l7policy = L7PolicyManager(self)
        self.l7rule = L7RuleManager(self)
        self.pool = PoolManager(self)
        self.member = MemberManager(self)
        self.health_monitor = HealthMonitorManager(self)


class LoadBalancerManager(driver_base.BaseLoadBalancerManager):

    @log_helpers.log_method_call
    def create(self, context, lb):
        self.successful_completion(context, lb)

    @log_helpers.log_method_call
    def update(self, context, old_lb, lb):
        if self.driver.workflow_exists(old_lb):
            self.driver.execute_workflow(
                context, self, lb, old_data_model=old_lb)
        else:
            self.successful_completion(context, lb)

    @log_helpers.log_method_call
    def delete(self, context, lb):
        if self.driver.workflow_exists(lb):
            self.driver.remove_workflow(
                context, self, lb)
        else:
            self.successful_completion(context, lb, delete=True)

    @log_helpers.log_method_call
    def refresh(self, context, lb):
        if lb.listeners and any(listener.default_pool and
            listener.default_pool.members for listener in lb.listeners):
            self.driver.execute_workflow(
                context, self, lb)
        else:
            self.successful_completion(context, lb)

    @log_helpers.log_method_call
    def stats(self, context, lb):
        if self.driver.workflow_exists(lb):
            return self.driver.get_stats(context, lb)
        else:
            self.successful_completion(context, lb)


class ListenerManager(driver_base.BaseListenerManager):

    @log_helpers.log_method_call
    def create(self, context, listener):
        if self.driver.workflow_exists(listener.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, listener)
        else:
            self.successful_completion(context, listener)

    @log_helpers.log_method_call
    def update(self, context, old_listener, listener):
        if self.driver.workflow_exists(old_listener.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, listener, old_data_model=old_listener)
        else:
            self.successful_completion(context, listener)

    @log_helpers.log_method_call
    def delete(self, context, listener):
        if self.driver.workflow_exists(listener.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, listener, delete=True)
        else:
            self.successful_completion(context, listener, delete=True)


class L7PolicyManager(driver_base.BaseL7PolicyManager):

    @log_helpers.log_method_call
    def create(self, context, policy):
        if self.driver.workflow_exists(policy.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, policy)
        else:
            self.successful_completion(context, policy)

    @log_helpers.log_method_call
    def update(self, context, old_policy, policy):
        if self.driver.workflow_exists(old_policy.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, policy, old_data_model=old_policy)
        else:
            self.successful_completion(context, policy)

    @log_helpers.log_method_call
    def delete(self, context, policy):
        if self.driver.workflow_exists(policy.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, policy, delete=True)
        else:
            self.successful_completion(context, policy, delete=True)


class L7RuleManager(driver_base.BaseL7RuleManager):

    @log_helpers.log_method_call
    def create(self, context, rule):
        if self.driver.workflow_exists(rule.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, rule)
        else:
            self.successful_completion(context, rule)

    @log_helpers.log_method_call
    def update(self, context, old_rule, rule):
        if self.driver.workflow_exists(old_rule.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, rule, old_data_model=old_rule)
        else:
            self.successful_completion(context, rule)

    @log_helpers.log_method_call
    def delete(self, context, rule):
        if self.driver.workflow_exists(rule.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, rule, delete=True)
        else:
            self.successful_completion(context, rule, delete=True)


class PoolManager(driver_base.BasePoolManager):

    @log_helpers.log_method_call
    def create(self, context, pool):
        if self.driver.workflow_exists(pool.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, pool)
        else:
            self.successful_completion(context, pool)

    @log_helpers.log_method_call
    def update(self, context, old_pool, pool):
        if self.driver.workflow_exists(old_pool.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, pool, old_data_model=old_pool)
        else:
            self.successful_completion(context, pool)

    @log_helpers.log_method_call
    def delete(self, context, pool):
        if self.driver.workflow_exists(pool.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, pool, delete=True)
        else:
            self.successful_completion(context, pool,
                                       delete=True)


class MemberManager(driver_base.BaseMemberManager):

    @log_helpers.log_method_call
    def create(self, context, member):
        self.driver.execute_workflow(
            context, self, member)

    @log_helpers.log_method_call
    def update(self, context, old_member, member):
        self.driver.execute_workflow(
            context, self, member, old_data_model=old_member)

    @log_helpers.log_method_call
    def delete(self, context, member):
        self.driver.execute_workflow(
            context, self, member,
            delete=True)


class HealthMonitorManager(driver_base.BaseHealthMonitorManager):

    @log_helpers.log_method_call
    def create(self, context, hm):
        if self.driver.workflow_exists(hm.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, hm)
        else:
            self.successful_completion(context, hm)

    @log_helpers.log_method_call
    def update(self, context, old_hm, hm):
        if self.driver.workflow_exists(old_hm.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, hm, old_data_model=old_hm)
        else:
            self.successful_completion(context, hm)

    @log_helpers.log_method_call
    def delete(self, context, hm):
        if self.driver.workflow_exists(hm.root_loadbalancer):
            self.driver.execute_workflow(
                context, self, hm, delete=True)
        else:
            self.successful_completion(context, hm,
                                       delete=True)
